/*****************************************************************************************
* File name: Check.c
* Purpose:  Acknowledges the Check Command.
*
* Description:	This Program Acknowledges the Reception of the Check Command
*				and sends the Checksum for a Memory Area in Case of Command Consistency.
*****************************************************************************************/

#ifdef SIMULATOR
#include "virtuosoSim.h"
#else // SIMULATOR
#include "node1.h"
#include "k_struct.h"
#endif
#include "l_hwmap.h"
#include "spulib.h"

/*=================================================================================================*/
/*Functions declared in this file
===================================================================================================*/

void Check(u4 buf[]);
u4 Crc(u4 *start, u4 length, u4 alignment);
u4 crc_core(u4 b, u4 S);

/*=================================================================================================*/
/*Global variables declared in this file
===================================================================================================*/

u4  ProcBuff[0x42000];	//a general purpose processing buffer

/*=================================================================================================*/
/*External function and variables
===================================================================================================*/

#ifndef SIMULATOR
extern void MEM_CopyPMToDM(u4 src, u4 *dest, u4 size);
#endif

/*================================================================================================*/

/*
This function handles the check commands received from DPU
*/
void Check(u4 buf[])
{
	struct FIFO_TRANSMIT_MSG fifoTransmitMsg;	//msg to communicate with the data tx task
	u4 address , address1, address2, memid, memtype, memblock, length;
	u4 i,address_ERAM, address_SMCS,checks, *temp_buffer;
	
	/*=============================================================================================*/
	
	/*Initialize the Temporary Buffer Address*/
	temp_buffer= ProcBuff;
	
	/* Read the Address1 */
	address1 = buf[0]&0x0000FFFF;
	
	/* The Memory ID */
	memid= address1>>8;
	
	/* Read the Address2 */
	address2 = buf[1]>>16;
	
	/* The Logical Address */
	address = ((address1&0xFF)<<16)|(address2&0xFFFF);
	address_ERAM=K_DMADDR_BASE_EXTDRAM+address;
	address_SMCS=K_DMADDR_BASE_SMCSDPRAM+address;
	
	/* The Data Length */
	length = buf[1]&0xFFFF;
	
	/* The Memory Type (Data Memory or Program Memory) */
	memtype = (memid >>4)&0x1;
	
	/* The Memory Block (RAM or Extended RAM) */
	memblock = memid &0xF;
	
	switch (memtype)
	{
	case 0: /* Case of the Program Memory */
		{
			/* Case of the EEPROM, Command will be Rejected */
			if (memblock==3)
			{
				fifoTransmitMsg.Header = K_NACK_CHECK | K_INVALID_MEM_ID;
				fifoTransmitMsg.Block1.NackParameter = memid;
			}
			else /* Case of the PRAM */
			{
				/* Positive Acknowledgment */	
				fifoTransmitMsg.Header = K_PACK_CHECK | address1;
				fifoTransmitMsg.Block1.AddressLength = (address2<<16)|length;
				
				/* The Data stored in ProcBuff */
#ifndef SIMULATOR
				MEM_CopyPMToDM(address,temp_buffer,length*3);
#endif
				/* Calculate the Checksum */
				checks=Crc(temp_buffer,length*3,32)<<16;
				fifoTransmitMsg.Checksum = checks;
			}
		}break;
		
	case 1: /* Case of the Data Memory */
		{
			/* Case of the EEPROM, Command will be Rejected */
			if (memblock==3)
			{
				fifoTransmitMsg.Header = K_NACK_CHECK | K_INVALID_MEM_ID;
				fifoTransmitMsg.Block1.NackParameter = memid;
			}
			
			else
			{
				/* Check the Address for DRAM ExtRAM and DPRAM */
				if ((address> K_DM_END_DRAM)||((address> K_DM_END_DPRAM)&&(memblock==4)))			
				{
					/* invalid Address*/
					fifoTransmitMsg.Header = K_NACK_CHECK | K_INVALID_ADDRESS;
					fifoTransmitMsg.Block1.NackParameter = ((address1&0xFF)<<16)|(address2&0xFFFF);
				}
				else /* Positive Acknowledgment */
					/* Check the Data Memory */
				{
					/* Calculate the Physical Addressin Case of Extended RAM or DPRAM*/
					if(memblock==2) 
					{
						address=address_ERAM;
					}
					else if(memblock==4) 
					{
						address=address_SMCS;
					}
					
					fifoTransmitMsg.Header = K_PACK_CHECK | address1;
					fifoTransmitMsg.Block1.AddressLength = (address2<<16)|length;
					
					/* The Data stored in ProcBuff */
					for (i=0;i<length;i++)
					{
						temp_buffer[i] = ((u4*)address)[i];
					}
					fifoTransmitMsg.Checksum=Crc((u4 *)temp_buffer,length*2,32)<<16;
				}
			}
		}break;
		
	default: /* Case of Invalid Memory Type */
		/* Invalid Memory ID (only the Memory Type is Mentioned Here)*/
		{
			fifoTransmitMsg.Header = K_NACK_CHECK | K_INVALID_MEM_ID;
			fifoTransmitMsg.Block1.NackParameter = memid;
		}break;
	}
	KS_FIFOPut(FIFO_TRANSMIT, &fifoTransmitMsg); /* Send the Response Packet to the DPU */
}

/********************************************************************************************
* Function name: Crc
*
* Purpose: Calculates the Checksum
*
* Description:	This Program Perform the Checksum of the Appropriate Memory Block.
*				It Uses the Cyclic Redundancy Check Code Described in the PS-ICD
*				Issue 2.Rev.0 of 12.07.2001   
*
* Syntax: unsigned int Crc(unsigned int *start, unsigned int length, unsigned int alignment)
*
* Input: 'address', 'Length' and 'Alignement'
*
* Output: Result in 16 Bits
*
* Return: Result in 16 Bits
*
* Files Accessed: None
*
* Creation date & author: 03.07.2001 by Roland Ottensamer
* Version, Update date & author: HPSPU-HLSW-OB-V1R0 04.01.2002 by  Ahmed Nabil Belbachir
*
********************************************************************************************/

u4 Crc(u4 *start, u4 length, u4 alignment)
{
/* returns a 16-bit CRC aligned in a 32-bit u4eger
length counts 16-bit words, alignment can be 32 or 48 
	!!! BIG ENDIAN VERSION !!! */
	
	register u4 S =0xFFFF, i;  
	
	/*=============================================================================================*/
	
	if(alignment == 32) /* Check in Data Memory */
    {
		for(i=0, length *= 2; i<length; i++)
		{
			S = crc_core((start[roldiv(i, 4)]) >> (24-(i%4)*8), S);
		}
	}
	else /* Check in Program Memory */		
	{
		for(i=0, length *= 3; i<length; i++)
		{
			S = crc_core((start[roldiv(i, 6)]) >> (40-(i%6)*8), S);
		}
	}
	return (S & 0xffff); /* The Calculated Checksum */
}


/********************************************************************************************
* Function name: crc_core
*
* Purpose: Calculates the Cyclic Redundancy Check
*
* Description:	This Function Calculates the Cyclic Redundancy Check of a Data
*				Block Using the Given Formula in PS-ICD Issue 2.0 Page 102   
*
* Syntax: unsigned int crc_core(register unsigned int address,register unsigned int result)
*
* Input: 'address', 'Result'(Result is initialized to 0xFFFF)
*
* Output: Result in 16 Bits
*
* Return: Result in 16 Bits
*
* Files Accessed: None
*
* Creation date & author: 03.07.2001 by Roland Ottensamer
* Version, Update date & author: HPSPU-HLSW-OB-V1R0 04.01.2002 by Ahmed Nabil Belbachir
*
********************************************************************************************/

u4 crc_core(register u4 b, register u4 S)
{   /* returns the 16-bit syndrome aligned in 32 bit 
	call with S = 0xffff at the first time! 
	b is 8-bit unsigned, aligned in 32 bit! */
	
    /* Declaration of Local Variables
	===============================================================================================*/
	
	register u4 i;
	
	/*=============================================================================================*/
	
	for(i=0; i<8; i++, b = b<<1)
	{
		S = (((b & 0x80) ^ ((S & 0x8000)>>8)) ? (S<<1) ^ 0x1021 : S<<1);
	}
	return (S & 0xffff);
}
