/***************************************************************************************
* File name: fill_out.c
* Purpose:  Builds the Compressed Entity.			 
*
* Description:	The Fill_out Program is Used to Prepare the Compressed
*				Entity and Build the PUS Packets. It sets the Packet Ready-to-
*				Transmit to the DataTransmission Function.
****************************************************************************************/

#ifdef SIMULATOR
#include "virtuosoSim.h"
#include "stdio.h"
#else // SIMULATOR
#include "node1.h"
#include "k_struct.h"
#endif
#include "spuAsw.h"
#include "spulib.h"

/*=================================================================================================*/
/*Global variables declared in this file
===================================================================================================*/

struct PROCESSING_RESULT Result;

/*=================================================================================================*/
/*External function and variables
===================================================================================================*/

extern COMPRESSED_DMC_HEADER_BUFFER CompressedDmcHeaderBuffer;
extern u4 AveragingParametersSummary;


/*================================================================================================*/
/*
A compressed entity is made of:
- 7 words of header
- the compressed DMC headers (if any) copied from CompressedDmcHeaderBuffer
- a compression/resampling/pre-processing parameter
- the compressed readouts copied from ProcBuf
- the raw channels copied from RawChannels

Once the compressed entity is built, it is split in blocks of 253 words made of:
- 3 words of header
- 250 words copied from the compressed entity

These blocks are copied in the circular buffer CBUF_TM and will be sent to DPU by datatx task.
*/
void BuildCompressedEntity(COMPRESSION_MODE mode)
{
	/* Declaration of Local Variables
	==============================================================================================*/
	
	struct FIFO_TRANSMIT_MSG fifoTransmitMsg;	//msg to communicate with the data tx task
	u4 blockHeader[3];
	u4 i;
	u4 checksum = 0;

	u4 compressedReadoutsSize;
	u4 compressedSienceDataSize;  
	u4 totalSize;
	u4 dmcHeaderSize = 0;
	u4 preProcessingParametersSize = 0;
	u4 nbBlocks;
	u4 blockCounter = 0;   

	u4 *compressedEntityBuffer = DramBuffer;	//the compressed entity will be stored in the DramBuffer before being sent to DPU (but, for code readability, we use another variable name)
	
	/*============================================================================================*/
		
	/* Read the size of the obtained data */
	compressedReadoutsSize = CompressedDataBuffer.SizeInWords;   

	/* Send the Compressed DMC Header if No Buffer Transmission Mode*/ 
	if (BufferTransmissionModeEnabled == K_NO)
	{
		dmcHeaderSize = CompressedDmcHeaderBuffer.SizeInWords;
		preProcessingParametersSize = 1;
	}

	/* Write the Size of Science Data
	( Compressed Readouts size +1( Offset Value) + RCX Data Size if exit) */ 
	compressedSienceDataSize = compressedReadoutsSize + RawChannels.SizeInWords + preProcessingParametersSize; 

	/* Prepare the Compressed Entity Header */
	compressedEntityBuffer[0] = mode.DetectorType;
	compressedEntityBuffer[1] = mode.PacketIndex;

	if (mode.DetectorType == K_SPECTRO)
	{
		compressedEntityBuffer[2] = ((Result.ReductionAlgo << 16) & 0xFFFF0000) | (RawChannelSelectionSpectro.NbSelectedRawChannels<<11)|(RawChannelSelectionSpectro.IndexOfFirstChannel);
	}
	else
	{
		compressedEntityBuffer[2] = ((Result.ReductionAlgo << 16) & 0xFFFF0000) | (RawChannelSelectionPhoto.NbSelectedRawChannels<<11)|(RawChannelSelectionPhoto.IndexOfFirstChannel);
	}

	compressedEntityBuffer[3] = ((K_SPU_HLSW_VERSION_ID<<8)| mode.CompressionMode)&0xFFFF;
	compressedEntityBuffer[4] = DetectorSelectionId;
	compressedEntityBuffer[5] = (RawChannels.SizeInWords <<16) | (dmcHeaderSize & 0xFFFF);
	compressedEntityBuffer[6] = compressedSienceDataSize;

	/* Copy the Compressed DMC Header into the Compressed Entity*/
    if (dmcHeaderSize!=0)
	{
		MemCopyDmToDm(CompressedDmcHeaderBuffer.Data, compressedEntityBuffer+7, dmcHeaderSize);
	}
	
	/*Copy the Preprocessing parameter in Case of Non-Buffer Transmission Mode */
	if (BufferTransmissionModeEnabled == K_NO)
	{
		if (mode.DetectorType == K_SPECTRO)
		{
			// the PrePro...Summary has "SortingAlgorithm" in the high 16 bits
			// and "LosslessCompressionAlgorithm" in the low 16 bits
			// We pack both into the high part to have space for the checksum
			compressedEntityBuffer[dmcHeaderSize+7] = (PreProcessingParametersSummary & 0xffff0000) << 8;
			compressedEntityBuffer[dmcHeaderSize+7] |= (PreProcessingParametersSummary & 0x0000ffff) << 16;
		}
		else
		{
			compressedEntityBuffer[dmcHeaderSize+7] = AveragingParametersSummary;
		}
	}

	/* Copy the Science Data (compressed readouts) in the Compressed Entity*/
	MemCopyDmToDm(CompressedDataBuffer.Data, (compressedEntityBuffer + 7 + dmcHeaderSize + preProcessingParametersSize), compressedReadoutsSize);

	/* Copy the Raw Channel Data in the Compressed Entity if Required*/
	if(RawChannels.SizeInWords != 0)
	{
		MemCopyDmToDm(RawChannels.CompressedData, (compressedEntityBuffer + 7 + dmcHeaderSize + preProcessingParametersSize + compressedReadoutsSize), RawChannels.SizeInWords);
	}

	/* Set the Packet Header According to the Data Type */
	fifoTransmitMsg.Header = K_SCIENCE_HEADER + (mode.DetectorType <<16); 
	blockHeader[0] = fifoTransmitMsg.Header;

	/* Total Size of Compressed Entity */
	totalSize = compressedSienceDataSize + dmcHeaderSize + 7;
	
	// compute a simple checksum (XOR) and include it at the end of the header in the averaging-preprocessing summary entry
	if ((preProcessingParametersSize > 0) && (SimulationDataParameters.ComputeChecksumOnScienceData != 0))
	{
		checksum = 0;
		for (i = 7; i < totalSize; i++)
		{
			checksum ^= compressedEntityBuffer[i];
		}
		compressedEntityBuffer[dmcHeaderSize+7] |= (checksum & 0xFFFF) ^ (checksum >> 16);
	}

	/*Number of TM Blocks. One Added Word to Avoid The Truncation Error */
	nbBlocks = roldiv(totalSize, K_SCIENCE_BLOCK_SIZE_IN_WORDS)+1;
	blockHeader[2] = nbBlocks;

#ifdef SIMULATOR
   printf("number of TMs %d \n", nbBlocks);
#endif

	/* Max Size of TM Packet */
	fifoTransmitMsg.Block1.Size = K_SCIENCE_BLOCK_SIZE_IN_WORDS + 3;	//3 words for the header of each block

	//default compression status
	Result.CompStatus = K_COMP_STATUS_CSW_RUNNING;

	/* Build TM Packets and Make Them Ready to Transmit */
	if (CompressionSoftwareRunning == K_YES) //check if the compression software has been stopped
	{
		/* Case of the Buffer Transmission Mode */
		/*================================================================================*/
		if(BufferTransmissionModeEnabled == K_YES)
		{
			fifoTransmitMsg.NbPacketsToTransmit = 0;
			while (blockCounter < nbBlocks)
			{
				//make sure there is enough room in the TM buffer.  If not, signal an error and abort the transmission of this Compressed Entity
				if (CircularBufferQuerySpace(K_CBUF_TM) <= (K_SCIENCE_BLOCK_SIZE_IN_WORDS + 3))
				{
					//clear the TM circular buffer that has been filled but will not be transmitted
					CircularBufferRemoveLastWrittenValues(K_CBUF_TM, fifoTransmitMsg.NbPacketsToTransmit * (K_SCIENCE_BLOCK_SIZE_IN_WORDS + 3));

					//signal the error
					PushDmcHeaderError(K_DMC_DPU_IS_TOO_SLOW);
					Result.CompStatus = K_COMP_STATUS_COMPRESSED_ENTITY_INCOMPLETE;
					break;
				}

				if (BufferTransmissionModeEnabled == K_YES) //check if the current buffer transmission has been interrupted or not
				{
					/*Write the Block Counter*/
					blockHeader[1] = blockCounter + 1;
				
					/* Store The TM packet in the Output Buffer */
					WriteInCircularBuffer(K_CBUF_TM, blockHeader, 3);
					WriteInCircularBuffer(K_CBUF_TM, compressedEntityBuffer + (K_SCIENCE_BLOCK_SIZE_IN_WORDS * blockCounter), K_SCIENCE_BLOCK_SIZE_IN_WORDS);

					/* Increment the Block Counter */
					blockCounter++;
					fifoTransmitMsg.NbPacketsToTransmit++;
					if(fifoTransmitMsg.NbPacketsToTransmit == K_MAX_NB_TM_PACKETS)
					{
						KS_FIFOPut(FIFO_TRANSMIT, &fifoTransmitMsg);
						fifoTransmitMsg.NbPacketsToTransmit=0;
						/*Delay Used in Buffer Transmission Mode to Fulfill the Burst Mode TM 
						(~260Kbits/s)*/
						KS_TaskSleep(2220);
					}
				}
				else
				{//a stop command has been received
					//and clear the TM circular buffer that has been filled but will not be transmitted
					CircularBufferRemoveLastWrittenValues(K_CBUF_TM, fifoTransmitMsg.NbPacketsToTransmit * (K_SCIENCE_BLOCK_SIZE_IN_WORDS + 3));

					//don't send the last data to DPU
					fifoTransmitMsg.NbPacketsToTransmit = 0;
					Result.CompStatus = K_COMP_STATUS_COMPRESSED_ENTITY_INCOMPLETE;
					break;
				}
			}		

			/* Send the Compressed Data in TM Packets to the DPU */
			if(fifoTransmitMsg.NbPacketsToTransmit != 0)
			{
				KS_FIFOPut(FIFO_TRANSMIT, &fifoTransmitMsg);	
				/*Delay Used in Buffer Transmission Mode to Fulfill the Burst Mode TM (~260Kbits/s)*/
				KS_TaskSleep(1800);
			}
			/* Set the Reduction Algorithm to None */
			Result.ReductionAlgo = 0xFF;
		}
		/*================================================================================*/
		/* Other Cases than the Buffer Transmission Mode */
		/*================================================================================*/
		else
		{	
			fifoTransmitMsg.NbPacketsToTransmit = nbBlocks;
			while (blockCounter < nbBlocks)
			{
				//make sure there is enough room in the TM buffer.  If not, signal an error and abort the transmission of this Compressed Entity
				if (CircularBufferQuerySpace(K_CBUF_TM) <= (K_SCIENCE_BLOCK_SIZE_IN_WORDS + 3))
				{
					//clear the TM circular buffer that has been filled but will not be transmitted
					CircularBufferRemoveLastWrittenValues(K_CBUF_TM, blockCounter * (K_SCIENCE_BLOCK_SIZE_IN_WORDS + 3));

					//signal the error
					Result.CompStatus = K_COMP_STATUS_COMPRESSED_ENTITY_INCOMPLETE;
					PushDmcHeaderError(K_DMC_DPU_IS_TOO_SLOW);
					fifoTransmitMsg.NbPacketsToTransmit = 0;
					break;
				}

				/*Write the Block Counter*/
				blockHeader[1] = blockCounter + 1;
				
				/* Store The TM packet in the Output Buffer */
				WriteInCircularBuffer(K_CBUF_TM, blockHeader, 3);
				WriteInCircularBuffer(K_CBUF_TM, compressedEntityBuffer + (K_SCIENCE_BLOCK_SIZE_IN_WORDS*blockCounter), K_SCIENCE_BLOCK_SIZE_IN_WORDS);
				
				/* Increment the Block Counter */
				blockCounter++;
			}

			/* Send the Compressed Data in TM Packets to the DPU */
			if(fifoTransmitMsg.NbPacketsToTransmit != 0)
			{
#ifdef SIMULATOR
				FILE* fp = fopen("ce.dat", "wb+");
				fwrite(compressedEntityBuffer, totalSize, 4, fp);
				fclose(fp);
#endif// SIMULATOR

				KS_FIFOPut(FIFO_TRANSMIT, &fifoTransmitMsg);
			}
				
			/* Case of Lossless or Transparent Mode*/
			if((mode.CompressionMode & 0x4)!=0)
			{
				Result.ReductionAlgo = 0xFF;
			}
		}
		
		/* Once the compressed entity has been transmitted to the datatx task, generate a hk packet with a description of the compressed entity */ 
		Result.ObsId = mode.ObsId;
		Result.PacketIndex = mode.PacketIndex;
		
		if (mode.DetectorType == K_SPECTRO)
		{
			Result.RawChannelIndex = RawChannelSelectionSpectro.IndexOfFirstChannel;
			if (DetectorConstantsSpectro.FixRawChannels != 0)
				RawChannelSelectionSpectro.IndexOfFirstChannel -= RawChannelSelectionSpectro.NbSelectedRawChannels;
		}
		else
		{
			Result.RawChannelIndex = RawChannelSelectionPhoto.IndexOfFirstChannel;
		}
		
		/* Send HK Packet After Compressed Entity */
		KS_SemaSignal(SEMA_HK);
	}	
	else
	{
		fifoTransmitMsg.NbPacketsToTransmit = 0;
	}
}
