/************************************************************************************ 
* File name: Llc.c 
* Purpose:  Performs Lossless Coding.                    
* 
* Description:   The Lossless Coding Program Depends on the Observing Mode Photometry
*				 (Llc_Phot) or Spectroscopy (Llc_Spec). The Arithmetic Encoding Algorithm
*				 is Used as Backend Lossless Compression Algorithm. 
****************************************************************************************/ 

#ifdef SIMULATOR 
#include "virtuosoSim.h"
#include "stdio.h"
#include "time.h"
#else // SIMULATOR 
#include "node1.h" 
#include "k_struct.h" 
#endif 

#include "spuAsw.h" 
#include "spulib.h" 
#include "pacscod.h"

/*=================================================================================================*/
/*Structures declared in this file
===================================================================================================*/

struct arimodel {
    u4 *freqtable; 
    u4 *cptable; // cumulative probability table 
    u4 *ncptable; // next cumulative probability table
    u4 *spillover; // swap buffer for spillover, i.e. words > 8 Bit   
    int spillctr;
    int probability;
};


/*=================================================================================================*/
/*Functions declared in this file
===================================================================================================*/

u4 ZeroRepetitionSuppression(u4 *source, u4 size, u4 *dest, u4 value);
unsigned int roldiv (unsigned int xx, unsigned int yy);
void ro_explode (u4 *src, u4 *dest, u4 srcSizeInWords);
void ari_findByteFreqsW (u4 *buffer, u4 nbytes, u4 *table);
int makeCumulative (u4 *table, u4 nrows, u4 *cumu);
void encodeSymbol (BIT_STREAM* bitStream, u4 cp, u4 ncp, u4 probability);
void flushEncoder (BIT_STREAM* bitStream);
int ariencW2 (u4 *source, u4 nbytes, u4 *dest, u4 *swap);
int getCompProbTable (u4 *model, u4 *dest);
int ari_compress_spu_x (u4 *source, u4 nbytes, u4 *dest, u4 *swap);
int arit_cod (u4 *source, u4 nbytes, u4 *dest, u4 *swap);
int rzip2 (u4 *source, u4 size, u4 *dest, u4 *swap, u4 range);
int vbwl_short (u4 *source, u4 words, u4 *dest, int blockSize);
u4 RZip32 (u4 *source, u4 sizeInBytes, u4 *dest, u4 *swap, u4 range);
void convert_byte_to_int(u4 *source, u4 *dest , u4 Size);
void convert_byte_to_intReverse(u4 *source, u4 *dest , u4 Size);
void compress_me(u4 dirsize, u4 buflen, u4 *buffer, u4 order);
void sort_me(u4 buflen, u4 *buffer, u4 order);
u4 new_srt(u4 *src, u4 size, u4 *dest, u4 order);
void CompressRawChannels(RAW_CHANNEL_DATA* rawChannels, u4 detectorType);
void ConvertSortedDataBuffer(u4 targetType);
void LosslessCompressionForPhoto(COMPRESSION_MODE mode);
void LosslessCompressionForSpectro(COMPRESSION_MODE mode);
void init_arimodel (struct arimodel *model, u4 *buffer, int symbols);
void encodeSym8k (BIT_STREAM* bitStream, u4 cp, u4 ncp, int probability);
void fm_flushEncoder (BIT_STREAM* bitStream);
void InitModelTable (u4 *destmodel, int tableid);
void update_fm_ari_model (u4 *buffer, u4 nwords, int symbols, int initval, struct arimodel *model);
int fm_encode_chunk (u4 *chunk, int chunksize, unsigned int *dest, struct arimodel *model);
int vbwl_midsize (u4 *source, int words, unsigned int *dest, int BS);
int fm_ariencode (u4 *source, u4 nwords, u4 *dest, u4 *swap, int modelId);

#ifdef SIMULATOR
void putnbits32(u4 value, u4 bitOffset, u4 nbits, u4 *dest);
void putnbits (u4 value, u4 offset, u4 nbits, u4 *dest);
#endif

/*=================================================================================================*/
/*Global variables declared in this file
===================================================================================================*/

struct BUFFER_DESCRIPTION CompressedDataBuffer;	//contains information about the compressed data buffer

/* new pacs_srt*/
u4 *out_now, point=0;
u4 recordsize=1;

/* Definitions for Arithmetic Coding */
#define ACROWS 256 
#define MAXFREQ 16380 /* Not 16383 Because of Word Alignment */ 

u4 ari_low = 0; 
u4 ari_high = 0xffff; 
u4 ari_underflow = 0; 
u4 ari_rctr = 0; 

/*=================================================================================================*/
/*External function and variables
===================================================================================================*/

#ifndef SIMULATOR
extern unsigned int roldiv (unsigned int dividend,unsigned int divisor);
extern void ro_explode (u4 *src, u4 *dest, u4 words);
#endif


/*=================================================================================================*/ 
/* Declaration of the Used Functions 
===================================================================================================*/ 
void LosslessCompressionForSpectro(COMPRESSION_MODE mode); 
void LosslessCompressionForPhoto(COMPRESSION_MODE mode); 
void CompressRawChannels(RAW_CHANNEL_DATA* rawChannels, u4 detectorType); 
int rzip2 (u4 *source, u4 size, u4 *dest, u4 *swap, u4 range); 
u4 RZip32 (u4 *source, u4 size, u4 *dest, u4 *swap, u4 range); 
int vbwl_short (u4 *source, u4 words, u4 *dest, int blockSize);

int fm_ariencode (u4 *source, u4 nwords, u4 *dest, u4 *swap, int modelId);


int arit_cod  (u4 *source, u4 nbytes, u4 *dest, u4 *swap); 
void encodeSymbol (BIT_STREAM* bitStream, u4 cp, u4 ncp, u4 probability);
extern void pacs_srt(u4 *inout, u4 length, u4 *indexlast, unsigned int order);


/*
This function suppresses zero repetition in bit streams.
*/
u4 ZeroRepetitionSuppression(u4 *source, u4 size, u4 *dest, u4 value)
{
	u4 ctr = 0; /* source buffer counter [32 bit words] */
	u4 dctr = 1; /* destination buffer counter [single bits] */
	u4 zerocounts=0;
	u4 key=0xD00000,key1=0x110000,key2=0x220000,key3=0x330000,key4=0x440000;
	u4 key5=0x550000,key6=0x660000,key7=0x770000,key8=0x880000,key9=0x990000;
	u4 key10=0xaa0000,key11=0xbb0000,key12=0xCB0000, key13=0xDC0000,key14=0xED0000,key15=0xF10000;
	u4 k1=1,k2=1,k3=1,k4=1,k5=1,k6=1,k7=1,k8=1,k9=1,k10=1,k11=1,k12=1,k13=1,k14=1,k15=1, mask=0xFFFF8000;
	
	/*============================================================================================*/
	
	/* Store the Original Size in 32 bit Words */
	dest[1] = size;
	
	for (ctr=0;ctr<size;ctr++)
	{
		if((source[ctr]&mask)==key1)
			k1=0;
		if((source[ctr]&mask)==key2)
			k2=0;
		if((source[ctr]&mask)==key3)
			k3=0;
		if((source[ctr]&mask)==key4)
			k4=0;
		if((source[ctr]&mask)==key5)
			k5=0;
		if((source[ctr]&mask)==key6)
			k6=0;
		if((source[ctr]&mask)==key7)
			k7=0;
		if((source[ctr]&mask)==key8)
			k8=0;
		if((source[ctr]&mask)==key9)
			k9=0;
		if((source[ctr]&mask)==key10)
			k10=0;
		if((source[ctr]&mask)==key11)
			k11=0;
		if((source[ctr]&mask)==key12)
			k12=0;
		if((source[ctr]&mask)==key13)
			k13=0;
		if((source[ctr]&mask)==key14)
			k14=0;
		if((source[ctr]&mask)==key15)
			k15=0;
	}
	if(k1==1)
		key=key1;
	else if (k2==1)
		key=key2;
	else if (k3==1)
		key=key3;
	else if (k4==1)
		key=key4;
	else if (k5==1)
		key=key5;
	else if (k6==1)
		key=key6;
	else if (k7==1)
		key=key7;
	else if (k8==1)
		key=key8;
	else if (k9==1)
		key=key9;
	else if (k10==1)
		key=key10;
	else if (k11==1)
		key=key11;
	else if (k12==1)
		key=key12;
	else if (k13==1)
		key=key13;
	else if (k14==1)
		key=key14;
	else if (k15==1)
		key=key15;
	
	dest[1] |=key;
	
	
	/* The Zero Suppression Procedure */
	for (ctr = 0; ctr < size;ctr++)
    {
		if(source[ctr]==value)
			zerocounts++;
		else
		{
			if(zerocounts>0)
			{ 
				dest[++dctr]=(key|zerocounts);
				zerocounts=0;
			}
			dest[++dctr]=source[ctr];
		}
	}
	
	if(zerocounts>0)
	{ 
		dest[++dctr]=(key|zerocounts);
	}
	dest[0] = dctr+1;
	
	return dest[0]; /* in 32 bit words */
}


int roldivs(int x, int y)
{
	int xsign = (x < 0);
	int ysign = (y < 0);
	int r;

	if (x < 0)
	{
		x = -x;
	}
	if (y < 0)
	{
		y = -y;
	}
	r = roldiv(x, y);
	if ((xsign ^ ysign) != 0)
	{
		r = -r;
	}
	return r;
}

#ifdef SIMULATOR
/*
This is the C implementation of roldiv.  The OBS uses the assembly version.
*/
unsigned int roldiv (unsigned int xx, unsigned int yy)
{
    unsigned int x = xx;
    unsigned int y = yy;
    int xsh = 0;
    int ysh = 0;
    
    int shift = 0;
    int result = 0;
	
    int ctr;
	
    if (y > x)
		return 0;
	
	if (y == 0)
		return 0xffffffff;
	
    // shift x and y to the left and remember the number of bits
    for (ctr = 0; !(x & 0x80000000); ctr++)
    {
		x <<= 1;
		xsh++;
    }
    for (ctr = 0; !(y & 0x80000000); ctr++)
    {
		y <<= 1;
		ysh++;
    }
	
    shift = ysh - xsh;
    
    x = xx;
    y = yy << shift;
	
    for ( ; shift; shift--)
    {
		if (x >= y)
		{
			x -= y;
			result += 1 << shift;
		}	
		y >>= 1;
    }
	
    if (x >= y)
		result += 1;
	
    return result;
}


/*
This function transforms a buffer of 32bits word into a buffer 4 times bigger where each
32bits word contains a byte.
*/
void ro_explode (u4 *src, u4 *dest, u4 srcSizeInWords)
{
	int i, pos;
	
	/*=============================================================================================*/ 
	/* Conversion of Data Alignment*/
	pos = srcSizeInWords*4;
	srcSizeInWords--;
	for (i = srcSizeInWords; i >= 0; i--)
	{      
		dest[--pos] = src[i] & 0xff;
		dest[--pos] = (src[i] >> 8) & 0xff;
		dest[--pos] = (src[i] >> 16) & 0xff;
		dest[--pos] = (src[i] >> 24) & 0xff;
	}
	return;
}
#endif


/********************************************************************************************
* Function name: ari_findByteFreqsW
*
* Purpose: Creates Frequency Table
*
* Description:	This Function Counts the Number of Times each Octet Value is Repeated and 
*				Stores them Numbers in a Table.
*
* Syntax: unsigned int ari_findByteFreqsW(unsigned int *address, unsigned int Size, u4 *Table)
*
* Input: Pointer to Data and Table Addresses and Size.
*
* Output: Table of Frequencies
*
* Return: Sum of All Frequencies
*
* Files Accessed: None
*
* Creation date & author: 28.01.2003 by Roland Ottensamer
*
* Version, Update date & author: HPSPU-HLSW-OB-V3R0 07.03.2003 by Ahmed Nabil Belbachir
*
********************************************************************************************/

void ari_findByteFreqsW (u4 *buffer, u4 nbytes, u4 *table)
{
	u4 ctr;  
	
	/* clean table */
	for (ctr=0; ctr < ACROWS; ctr++)
	{
		table[ctr] = 0;
	}
		  
	/* find freqs */
	for (ctr=0; ctr < nbytes; ctr++)    
	{
		table[buffer[ctr]]++;
	}
}

/********************************************************************************************
* Function name: makeCumulative
*
* Purpose: Creates Probability Table 
*
* Description:	This Function Uses the Frequency Table to Create the Cumulative Probability
*				for Each Octet Value.
*
* Syntax: int makeCumulative(unsigned int *Table, unsigned int Size, u4 *Cumulative_Table)
*
* Input: Pointer to Table and Cumulative_Table Addresses and Size.
*
* Output: Table of Probabilites
*
* Return: Sum of All Probabilities
*
* Files Accessed: None
*
* Creation date & author: 28.01.2003 by Roland Ottensamer
*
* Version, Update date & author: HPSPU-HLSW-OB-V3R0 07.03.2003 by Ahmed Nabil Belbachir
*
********************************************************************************************/

int makeCumulative (u4 *table, u4 nrows, u4 *cumu) 
{ 
	u4 ctr; 
	
	/* clean table for the "cumulative probabilities"*/ 
	for (ctr=0; ctr < nrows; ctr++) 
	{
		cumu[ctr] = 0; 
	}
	
	/* the new table is +1 in size !! */
	for (ctr=0; ctr < nrows; ctr++)     
	{
		cumu[ctr+1] = cumu[ctr] + table[ctr]; 
	}
    
	return cumu[nrows]; 
} 

/********************************************************************************************
* Function name: encodeSymbol and flushEncoder
*
* Purpose: Symbols Encoding
*
* Description:	These Functions uses the Cumulative Probability for a given symbol(octet)
*				to add it to the encoded output. flushEncoder is used to Encode symbols at 
*				Buffer Border.
*
* Syntax: void encodeSymbol(BIT_STREAM* bitStream, unsigned int cp, unsigned int ncp ,u4 probability)
*		  void flushEncoder (BIT_STREAM* bitStream) 
*
* Input: Destination Address, Cummulative probability, Pointer to Cummulative Probability 
*		 and Sum of Probabilities
* Output: Encoded data
*
* Return: None
*
* Files Accessed: None
*
* Creation date & author: 28.01.2003 by Roland Ottensamer
*
* Version, Update date & author: HPSPU-HLSW-OB-V3R0 07.03.2003 by Ahmed Nabil Belbachir
*
********************************************************************************************/

void encodeSymbol (BIT_STREAM* bitStream, u4 cp, u4 ncp, u4 probability) 
{ 
	u4 width,ari_high_15Bit; 
	
	/*=============================================================================================*/ 
	
	/* Calculate the new interval */ 
	ari_low &= 0xffff; 
	ari_high &= 0xffff;   
	width = (ari_high - ari_low) + 1; 
	
	/* L + Pni * (H - L)*/
	ari_high = (ari_low + roldiv((ncp * width), probability) -1)&0xFFFF;  
	//	ari_high = (ari_low + pacs_div((ncp * width), probability) -1)&0xFFFF;  
	
	/* L + Pci * (H - L) */
	ari_low = (ari_low + roldiv((cp * width), probability))&0xFFFF;  
	//	ari_low = (ari_low + pacs_div((cp * width), probability))&0xFFFF;  
	
	/* L + Pni * (H - L)*/
	//	ari_high = (ari_low + (ncp * width)/ probability -1)&0xFFFF;  
	
	/* L + Pci * (H - L) */
	//	ari_low = (ari_low + (cp * width)/ probability)&0xFFFF;  
	
	/* Main Loop*/ 
	for ( ; ; ) 
	{ 
		/* Write Out Equal Bits */ 
		if ((ari_high & 0x8000) == (ari_low & 0x8000)) 
		{ 
			ari_high_15Bit =(ari_high>>15)&0x1;
			if(ari_high_15Bit)
			{
				PushBitTrue(*bitStream);
				//MEM32_SET_BIT(dest, ari_wctr++);
			}
			else
			{
				PushBitFalse(*bitStream);
				//MEM32_CLR_BIT(dest, ari_wctr++);
			}
			
			while (ari_underflow > 0) 
			{ 
				if(ari_high_15Bit)
				{
					PushBitFalse(*bitStream);
					//MEM32_CLR_BIT(dest, ari_wctr++);
				}
				else
				{
					PushBitTrue(*bitStream);
					//MEM32_SET_BIT(dest, ari_wctr++);
				}
				ari_underflow--; 
			} 
		} 
		/* underflow coming up, because <> and the 2nd bits are just one apart */
		else if ((ari_low & (~ari_high)) & 0x4000) 
		{ 
			ari_underflow++; 
			ari_low &= 0x3fff; 
			ari_high |= 0x4000; 
		} 
		else 
		{
			return ; 
		}
		
		ari_low <<= 1; 
		ari_low &= 0xffff; 
		ari_high <<= 1; 
		ari_high |= 1; 
		ari_high &= 0xffff; 
	} 
} 

void flushEncoder (BIT_STREAM* bitStream) 
{ 
	u4 ari_low_14Bit=(ari_low>>14)&0x1;
	
	if(ari_low_14Bit)
	{
		PushBitTrue(*bitStream);
		//MEM32_SET_BIT(dest, ari_wctr++);
	}
	else
	{
		PushBitFalse(*bitStream);
		//MEM32_CLR_BIT(dest, ari_wctr++);
	}
	
	ari_underflow++; 
	while (ari_underflow-- > 0)      
	{
		if(ari_low_14Bit)
		{
			PushBitFalse(*bitStream);
			//MEM32_CLR_BIT(dest, ari_wctr++);
		}
		else
		{
			PushBitTrue(*bitStream);
			//MEM32_SET_BIT(dest, ari_wctr++);
		}
	}
	return;
} 

/********************************************************************************************
* Function name: ariencW2
*
* Purpose: Aritmetic Coding
*
* Description:	This Function performs the Arithmetic Coding of the Data.
*
* Syntax: int ariencW2(unsigned int *source, unsigned int Size, unsigned int *dest ,u4 *swap)
*
* Input: Source, Destination, and Swap Addresses and Size of The Data
*
* Output: Encoded data
*
* Return: Size of Encoded Data in Bits
*
* Files Accessed: None
*
* Creation date & author: 28.01.2003 by Roland Ottensamer
*
* Version, Update date & author: HPSPU-HLSW-OB-V3R0 07.03.2003 by Ahmed Nabil Belbachir
*
********************************************************************************************/

int ariencW2 (u4 *source, u4 nbytes, u4 *dest, u4 *swap)
{
	u4 ctr,probability;
	u4 *freqtable = swap; 
	u4 *cptable = freqtable + ACROWS;	/* Cumulative Probability Table */ 
	u4 *ncptable = cptable + 1;			/* Next Cumulative Probability Table */
	u4 symbol, cp, ncp;
    
	BIT_STREAM bitStream;
	DeclareBitStream(bitStream, dest);
	
	/*=============================================================================================*/ 
	
	/* Make Probability Model */
	ari_findByteFreqsW (source, nbytes, freqtable);
	probability = makeCumulative (freqtable, ACROWS, cptable); /* should be = ACROWS */
	
	
	/* Initialization of the Parameters */
	ari_low = 0;
	ari_high = 0xffff;
	ari_underflow = 0;
	//	ari_wctr = 0;
	
	for (ctr=0; ctr < nbytes; ctr++)
    {
		/* Read Next Symbol */
		symbol = source[ctr];
		
		/* Look it Up in the Tables*/
		cp = cptable[symbol];
		ncp = ncptable[symbol];
		
		/* Encode the Symbol */
		encodeSymbol (&bitStream, cp, ncp, probability);
    }
	
	/* Encode the Rest */
	flushEncoder (&bitStream);
	
	return bitStream.CurrentBitPos; /* Number of Written Bits */
}

/********************************************************************************************
* Function name: getCompProbTable
*
* Purpose: Store the Probability Model
*
* Description:	This Function Writes the Probability Tables with the Encoded Data.
*
* Syntax: int getCompProbTable(unsigned int *model, unsigned int *dest)
*
* Input: Probability Table and Destination Address
*
* Output: Probability Table Transfered to Destination Address
*
* Return: Size of Probability Table in Bytes
*
* Files Accessed: None
*
* Creation date & author: 28.01.2003 by Roland Ottensamer
*
* Version, Update date & author: HPSPU-HLSW-OB-V3R0 07.03.2003 by Ahmed Nabil Belbachir
*
********************************************************************************************/

int getCompProbTable (u4 *model, u4 *dest) 
{ 
	u4 newsize, par_rows=(ACROWS>>1); 
	
	/*=============================================================================================*/ 
	/* Make it Half Size First */ 
	Put2ShortsIn1Word (model, model, par_rows); // /2 is needed! 
	
	for (newsize=0; newsize < par_rows; newsize++) 
	{
		dest[newsize] = model[newsize]; 
	}
	newsize *= 4; /* to get Bytes */ 
	return newsize; 
} 

/********************************************************************************************
* Function name: ari_compress_spu_x
*
* Purpose: Performs Arithmetic Coding for one Buffer
*
* Description:	This Function Call the Sequence Required to Perform the Arithmetic Coding for
*				one Buffer of Data.
*
* Syntax: int ari_compress_spu_x(unsigned int *source, unsigned int Size, unsigned int *dest, unsigned int *swap)
*
* Input: Source, Destination, and Swap Addresses and Size of The Data
*
* Output: Compressed Data at dest Address
*
* Return: Size of Compressed Data
*
* Files Accessed: None
*
* Creation date & author: 28.01.2003 by Roland Ottensamer
*
* Version, Update date & author: HPSPU-HLSW-OB-V3R0 07.03.2003 by Ahmed Nabil Belbachir
*
********************************************************************************************/

int ari_compress_spu_x (u4 *source, u4 nbytes, u4 *dest, u4 *swap)
{
	int newsize, cmpmodelsize;
	u4 *newdest;
	
	/*=============================================================================================*/ 
	/* Storage of Data Size and Initialization of the Parameters*/
	newdest = dest+1; /* left out for datasize | original*/
	
	ari_low = 0;
	ari_high = 0xffff;
	ari_underflow = 0;
	//ari_wctr = 0;
	ari_rctr = 0;
    
	/* Performs Compression of the Buffer */
	newsize = (ariencW2 (source, nbytes, newdest, swap)+7) >> 3;
	dest[0] = newsize << 16; /* on the left is the cmp data size in Bytes */
	cmpmodelsize = getCompProbTable (swap, newdest+((newsize+3) >> 2)); /* align in 4B */
	dest[0] |= nbytes; /* on the right is the original size in B */
	newsize = 4 + (((newsize+3)>>2) + ((cmpmodelsize+3)>>2)) * 4; /* 4B alignment */
	return newsize;
}

/********************************************************************************** 
* Function name: arit_cod 
* 
* Purpose: Compress the Data Lossless 
* 
* Description:   This Function Uses the Arithmetic Coding Algorithm to Perform 
*                the Lossless Coding of the Data. This Main Function of the
*				 Arithmetic Coding Could Process More than one Buffer 
* 
* Syntax: unsigned int arit_cod(unsigned int *start,unsigned int size,unsigned int *destunsigned int *swap) 
* 
* Input: start and swap addresses of the buffers and size of the Data 
* 
* Output: Size 
* 
* Return: Compressed Data size in Bytes 
* 
* Files Accessed: None 
* 
* Creation date & author: 02.02.2003 by Roland Ottensamer 
* 
* Version, Update date & author: HPSPU-HLSW-OB-V3R0 07.03.2003 by Ahmed Nabil Belbachir 
*
****************************************************************************************/ 

int arit_cod (u4 *source, u4 nbytes, u4 *dest, u4 *swap)
{
	u4 newsize=0;
	int ctr, remaining=nbytes, left; 
	u4 *newdest = dest;
	
	/*=============================================================================================*/ 
	
	/* Store Original (Total) Size */
	newdest[0] = nbytes; 
	
	/* Data and Model Size will be Stored in Dest[1] */ 
	newdest++; 
	
	/* Convert Buffer from 4-bytes Alignment to 1-Byte Alignment*/
	//explodeBuffer (source, ((nbytes+3)>>2));
	ro_explode (source, source, ((nbytes+3)>>2)); 
	
	/* Compression of the Data */
	for (ctr=0; remaining > 0; ctr++)
    {
		if (remaining > MAXFREQ)
		{
			left = MAXFREQ;
		}
		else
		{
			left = remaining;
		}
		
		/* Call Arithmetic Coding for One Buffer */
		newsize += ari_compress_spu_x (source+ctr*MAXFREQ, left, newdest + ((newsize+3) >> 2), swap);
		
		remaining -= MAXFREQ;
    }
	return newsize+4; /* +4 for the Original Size Stored in the Header */
}


/********************************************************************************** 
* Function name: rzip2
* 
* Purpose: Compress the Data Lossless 
* 
* Description: same as RZip32, but answers and ranges are coded at the end, 
*              leading to a better iteration power.
* 
* Creation date & author: 07.07.2006 by Roland Ottensamer 
* 
* Version, Update date & author: HPSPU-HLSW-OB-V12R7 9.7.2006 by Roland Ottensamer 
*
****************************************************************************************/ 
//
int rzip2 (u4 *source, u4 size, u4 *dest, u4 *swap, u4 range)
{
    // general top-level variables
    u4 osizew = (size+3)/4; // the not-compressed source buffer size in 4B-words
    u4 newsize; // the compressed size in Bytes
    u4 maxoff; // the maximum offset possible with the current range setting
    u4 symbol=0; // the protagonist of the whole story
    u4 ctr; // main counter, counting thru the words
    
    // inner loop variables doing the entry search
    u4 sctr; // search counter, used to find the next equal symbol
    u4 uctr; // counting the number of symbols left out during the search
    u4 lastpos; // holds the positions where symbols occur
    
    u4 i; // all-purpose counter
    
    // pointers and their counters we need 
    u4 *syms = dest + 2;
    u4 syms_ctr = 0; // counts words in syms (dest) buffer
	
	BIT_STREAM answers;
	BIT_STREAM rangebuf;
	
    u4 *contract = syms + osizew; // safe init 
    u4 contractsize = 0; // safe init
	
	DeclareBitStream(answers, syms + osizew);
	DeclareBitStream(rangebuf, answers.Buffer + ((osizew+31) >> 5));
	
	
    // maximum offset according to the range
    maxoff = 0xffffffff >> (32-range); 
	
    // important: clean the alpha channel first
    for (i=0; i < osizew; i++)
		swap[i] = 0; 
    
    // general loop
    for (ctr=0; ctr < osizew; ctr++)
    {
		// find a fresh symbol
		for ( ; (swap[ctr]==1) & (ctr<osizew); ctr++);	    
		
		if (ctr < osizew)
		{
			symbol = source[ctr]; // get it
			swap[ctr] = 1; // set it to used state 
			
			// encode the symbol
			syms[syms_ctr++] = symbol;
		}      
		
		// search the buffer from the symbol onward
		for (uctr=0, sctr=ctr+1, lastpos=ctr; (sctr < osizew) & ((sctr-lastpos-uctr) < maxoff); sctr++)
		{
			// if the current entry has been coded before, keep this in mind
			uctr += (swap[sctr]==1);
			
			// find and code an offset
			if ((symbol==source[sctr]) & (swap[sctr]==0)) 
			{
				// first code YES
				PushBitTrue(answers);
				
				// then code the reduced offset
				// that is: (current search position) - (where the last occurrence was) - 
				//          - (number of already before coded entries) - 1 (0 means neighbour)
				
				PushBits(rangebuf, sctr-lastpos-uctr-1, range);
				//putnbits (sctr-lastpos-uctr-1, rangebuf_ctr, range, (unsigned int*)rangebuf);
				//rangebuf_ctr += range;
				
				// now we set the current finding to used state
				// printf ("swapctr2 is: %d\n", sctr);
				swap[sctr] = 1;
				
				// set the last position of occurrency to the current one
				lastpos = sctr;
				
				// reset the number of not countable entries
				uctr = 0;
			}
		}
		
		// we come here if no enty was found for the current symbol
		if (ctr < osizew)
		{
			//	MEM32_CLR_BIT(answers, answers_ctr++);
			PushBitFalse(answers);
		}
		
		// go back until the end of the file is reached
    }
	
	PadBitStreamWithZeroes(answers);
	
    // write the header
    dest[0] = size; // original size in B
    
    // enter the number of coded symbols
    dest[1] = syms_ctr;
	
    // contract the buffers
    contract = dest + 2 + syms_ctr;
    //contractsize = (osizew+31)/32 + (rangebuf_ctr+31)/32; // answers and rangebuf in words
	contractsize = (osizew+31)/32 + (rangebuf.CurrentBitPos+31)/32; // answers and rangebuf in words
	
    for (i = 0; i < contractsize; i++)
	{
		contract[i] = answers.Buffer[i];
	}
    
    // and the total compressed size in B is now 
    newsize = 8 + syms_ctr*4 + contractsize*4; 
    
    return newsize;
}


/********************************************************************************** 
* Function name: vbwl_short 
* 
* Purpose: Compress the Data Lossless 
* 
* Description:   This is a Variable Block Word Length encoder. 
* 
* Syntax: int vbwl_short (u4 *start, u4 size, u4 *dest, int blockSize) 
* 
* Input: start and dest addresses of the buffers, size of the Data in words
*        and the blocksize (meaningful values: 1-n)  
*         
* Note: input data must be mapped to positive already, otherwise this is inefficient!
* 
* Output: Size in words (and the compressed buffer in dest, of course)
* 
* Return: Compressed Data size in Words
* 
* Files Accessed: None 
* 
* Creation date & author: 8.7.2006 by Roland Ottensamer 
* 
* Version, Update date & author: HPSPU-HLSW-OB-V12R7 9.7.2006 by Roland Ottensamer 
*
****************************************************************************************/ 
// 
#define SMALLCODE 3
#define BIGCODE 7

int vbwl_short (u4 *source, u4 words, u4 *dest, int blockSize)
{
    u4 ctr, bctr;
    u4 bits, width;
    //int outbits = 32; // keep track of the output bits; we start at dest[1]
	BIT_STREAM bitStream;
	
	DeclareBitStream(bitStream, dest + 1);	//the bit stream start at 1, the first value is reserved for the size
	
    // main loop counts through the source words
    for (ctr=0; ctr < words; ctr++)
    {
		// block-loop, we count through the words of a block
		for (width=0, bctr=ctr; (bctr < ctr+blockSize) & (bctr < words); bctr++)
		{
			// determine used bits of current word
			for (bits=1; (source[bctr] >> bits) > 0; bits++); // start with 1, otherwise the word 0 would use 0 bit
			
			width = bits > width ? bits : width;
		}
		// now we know width = the number of bits to encode the block
		
		// first code the width
		
		// decide whether to use 3 or (3+)4 bits for encoding the width
		if (width >= 16)
		{
			PushBits(bitStream, 0, BIGCODE);
			width = 16;
		}
		else if (width >= 8)
		{
			PushBits(bitStream, width, BIGCODE);
		}
		else
		{
			PushBits(bitStream, width, SMALLCODE);
		}
		
		// now code the words of the block
		for (bctr=ctr; (ctr < bctr+blockSize) & (ctr < words); ctr++)
		{	
			PushBits(bitStream, source[ctr], width);
		}
		ctr--;
    }
	
	PadBitStreamWithZeroes(bitStream);
	
    // store the original size
    dest[0] = words;
	
    // return the size in words, rounding up
    return (bitStream.CurrentBitPos + 32 + 31 )/32;  //the +32 is there to take the dest[0] into account, the +31 is there to round at upper word size
}


/***************************************************************************** 
* Function name: Rzip32 
* 
* Purpose: Compress the Data Lossless 
* 
* Description:   This Function Uses the RZIP Algorithm to Perform 
*                                the Lossless Coding of the Data. 
* 
* Syntax:u4 RZip32 (u4 *source, u4 size, u4 *dest, u4 *swap, u4 range) 
* 
* 
* Input: start  Dest and swap addresses of the buffers and size of the Data and the range 
* 
* Output: Size 
* 
* Return: Compressed Data size in Bytes 
* 
* Files Accessed: None 
* 
* Creation date & author: 22.12.2001 by Roland Ottensamer 
* 
* Version, Update date & author: HPSPU-HLSW-OB-V10R1 01.06.2004 by Ahmed Nabil Belbachir 
********************************************************************************/ 

u4 RZip32 (u4 *source, u4 sizeInBytes, u4 *dest, u4 *swap, u4 range) 
{ 
	u4 sizeInWords = (sizeInBytes+3)>>2; /* Original Data Size in 4B-Words */ 
	u4 newSizeInBytes; /* Compressed Size in Bytes */
	
	u4 maxoff; /* Maximum Offset Possible with the Current Range Setting */ 
	u4 symbol, ctr,i; /* Main Words Counter */ 
	
	/* Inner Loop Variables Doing the Entry Search */ 
	u4 sctr; /* Search Counter,to Find Next Equal Symbol */
	u4 uctr; /* Counting the Number of Symbols Left out During the Search */ 
	u4 wctr; /* Counts the Destination Bits */ 
	u4 lastpos; /* Holds the Positions where Symbols Occur */ 
	
	BIT_STREAM bitStream;
	DeclareBitStream(bitStream, dest + 1); //the bit stream start at 1 in the destination buffer, the first value is reserved for the size
	
	/*=============================================================================================*/ 
	
	/* Maximum Offset According to the Range */ 
	maxoff = 0xffffffff >> (32-range); 
	
	/* Clean the Alpha Channel */ 
	for (i=0; i < sizeInWords; i++) 
		swap[i] = 0; 
	
	/* Main Loop */ 
	for (ctr=0, wctr=32; ctr < sizeInWords; ctr++) //we start filling the bit stream at pos 32 because, later, the 32 first bits will contain the size
	{ 
		/* Find a Fresh Symbol */ 
		while ((swap[ctr]==1) & (ctr < sizeInWords)) 
		{
			ctr++; 
		}
		/* Read Symbol */ 
		symbol = source[ctr];
		
		/* Set it to Used State */ 
		swap[ctr] = 1; 
		      
		/* Encode the Symbol */ 
		PushBits(bitStream, symbol, 32);
		//putnbits (symbol, wctr, 32, dest);
		//wctr += 32; 
		      
		/* Search the Buffer from the Symbol Onwards */
		for (uctr=0, sctr=ctr+1, lastpos=ctr; (sctr < sizeInWords) & ((sctr-lastpos-uctr) < maxoff); sctr++) 
		{ 
			/* if the Current Entry has been Already Coded Before, Keep this in Mind */ 
			uctr += (swap[sctr]==1); 
			
			/* Find and Code an Offset */ 
			if (symbol==source[sctr]) 
			{ 
				/* First Code YES */ 
				PushBitTrue(bitStream);
				//MEM32_SET_BIT(dest,wctr); 
				//wctr++; 
				
				/* Code the Reduced Offset 
				That is: (Current Search Position) - (Where the Last Occurrence Was) - 
				- (Nb of Already Before Coded Entries) - 1 (0 Means Neighbour) */
				
				PushBits(bitStream, (sctr-lastpos-uctr-1), range);
				//putnbits ((sctr-lastpos-uctr-1), wctr, range, dest);
				//wctr += range; 
				
				/* Set the Current Finding to Used State */
				swap[sctr] = 1; 
				
				/* Set the Last Position of Occurrency to the Current One */ 
				lastpos = sctr; 
				
				/* Reset the Number of not Countable Entries */
				uctr = 0; 
			} 
		} 
		
		/* We Come Here if no Entry was Found for the Current Symbol */ 
		if (ctr < sizeInWords) 
		{	 
			/* We Code a NO */ 
			PushBitFalse(bitStream);
			//MEM32_CLR_BIT(dest,wctr); 
			//wctr++; 
		} 
		/* Go Back Until the End of the File is Reached */ 
	} 
	
	PadBitStreamWithZeroes(bitStream);
	
	/* Compressed Size in Bytes */
	newSizeInBytes = (bitStream.CurrentBitPos+7 + 32)>>3; //the 7 is there to round to upper byte size, the 32 is there to take the first word into account (the size of the buffer)
	
	/* Write the Header */
	dest[0] = sizeInWords * 4;
	//putnbits(size, 32, 0, dest);
	
	return newSizeInBytes; 
}

void convert_byte_to_int(u4 *source, u4 *dest , u4 Size)
{
	u4 l,k=0;
	
	/*=============================================================================================*/
	
	   for(l=0;l<Size;l++)
	   {
		   dest[l] = ((source[k]<<24)&0xFF000000)|((source[1+k]<<16)&0xFF0000)|((source[2+k]<<8)&0xFF00)|(source[3+k]&0xFF);
		   k+=4;
	   }
}

void convert_byte_to_intReverse(u4 *source, u4 *dest , u4 Size)
{
	u4 l,k=0;
	
	/*=============================================================================================*/
	
	   for(l=0;l<Size;l++)
	   {
		   dest[l] = ((source[3+k]<<24)&0xFF000000)|((source[2+k]<<16)&0xFF0000)|((source[1+k]<<8)&0xFF00)|(source[k]&0xFF);
		   k+=4;
	   }
}

//===========================================================
void compress_me(u4 dirsize, u4 buflen, u4 *buffer, u4 order)
{   
	u4 indexlast, ordi;
	u4 i=0, begin,ch;
	
#ifdef SIMULATOR
	int t0,t1,t2,t3,t4;
#endif
	
#ifdef SIMULATOR
	t0=clock();
#endif
	
	
	if((order<3)||(order>4))
	{
		Result.LlcError |= 0x8000;
		order=3;
	}
	
	pacs_srt(buffer,buflen,&indexlast,order);
#ifdef SIMULATOR
	t1= clock()-t0;
	t0=clock();
#endif
	
	ordi=(indexlast>>16)|((indexlast<<16)&0xFF0000)|(indexlast&0xFF00);
    out_now[point++]=ordi|(order<<24);
	
    initmodel(&m, dirsize+5, &recordsize);
	{
		buffer[buflen]= ~buffer[buflen-1];
		{
			ch = buffer[i++];
			while (buffer[i]==ch)
				i++;
			pacs_encode(&m, ch, i);
		}
		fixafterfirst(&m);
#ifdef SIMULATOR
		t2= clock()-t0;
		t0=clock();
#endif
		
		while (i<buflen)
		{   
			begin = i;
			ch = buffer[i++];
			while (buffer[i]==ch)
				i++;
			pacs_encode(&m, ch, i-begin);
		}
	}
#ifdef SIMULATOR
	t3= clock()-t0;
	t0=clock();
#endif
	
    deletemodel(&m);
#ifdef SIMULATOR
	t4= clock()-t0;
#endif
	
}
/*=================================================================================================*/ 
void sort_me(u4 buflen, u4 *buffer, u4 order)
{   
	u4 indexlast, ordi;
	pacs_srt(buffer,buflen,&indexlast,order);
	
	ordi=(indexlast>>16)|((indexlast<<16)&0xFF0000)|(indexlast&0xFF00);
    out_now[point++]=ordi|(order<<24);
	buflen++;
	while(point<buflen)
	{
		out_now[point++]=*(buffer++);
	}
}

/*=================================================================================================*/ 
u4 new_srt(u4 *src, u4 size, u4 *dest, u4 order) 
{ 
	u4 newsiz;
	
	point=0;
	
	//explodeBuffer (buf, size);
	ro_explode(src, src, size);
	
	out_now=dest;
    dest[point++]=size*4;
	
	/* Sort and Code*/
	compress_me(6, size*4, src, order);
    out_now[point]=out_now[point+1]=out_now[point+2]=0;
	newsiz= (point+1)>>2;
	
	convert_byte_to_int(dest+2,dest+2,newsiz);
	
	return newsiz;
}
/*=================================================================================================*/ 

/********************************************************************************************
* Function name: CompressRawChannels

 *
 * Purpose: Compress Data from the Selected Raw Channel 
 *
 * Description:	This Function is used for the Lossless Compression of the Raw Channel
 *               Data in Order to Downstream the Downlink.
 *
 * Input: Address and Size of Data Buffer.
 *
 * Output: The Results are Re-stored in the dest
 *
 * Return: None
 *
 * Files Accessed: None
 *
 * Creation date & author: 01.07.2003 by Ahmed Nabil Belbachir
 *
 * Version, Update date & author: HPSPU-HLSW-OB-V12R0 30.11.2005 by Ahmed Nabil Belbachir
 *
********************************************************************************************/

void CompressRawChannels(RAW_CHANNEL_DATA* rawChannels, u4 detectorType) 
{ 
    u4 newSize = 0;
	u4 *swap = DramBuffer+1;; 
	
	/*============================================================================================*/ 
    /* RZip Algorithm is used*/ 
    /*=============================================================================================*/ 
	
	if(detectorType == K_SPECTRO) /* Spectroscopy */
	{
		newSize = (arit_cod  (rawChannels->Data, rawChannels->SizeInWords*4, rawChannels->CompressedData, swap)+3)>>2;
	}
	else /* Photometry*/
	{
		newSize = (RZip32  (rawChannels->Data, rawChannels->SizeInWords*4, rawChannels->CompressedData, swap,5)+3)>>2;
		rawChannels->CompressedData[0] |= 0x50000000 ;
	}
	
	/* Write the Size of the Coded Data*/ 
	rawChannels->SizeInWords = newSize;
} 

/*
This function is used to convert the SortedDataBuffer into the type required by a lossless compression algorithm
*/
void ConvertSortedDataBuffer(u4 targetType)
{
	if (targetType == SortedDataBuffer.Type)
	{
		return;
	}
	else if (targetType == K_DATA_TYPE_1_SHORT_PER_WORD)
	{
		if (SortedDataBuffer.Type == K_DATA_TYPE_2_SHORTS_PER_WORD)
		{
			Put2ShortsIn2Words(SortedDataBuffer.Data, DramBuffer, SortedDataBuffer.SizeInWords);
			SortedDataBuffer.SizeInWords *= 2;
			MemCopyDmToDm(DramBuffer, SortedDataBuffer.Data, SortedDataBuffer.SizeInWords);
			SortedDataBuffer.Type = K_DATA_TYPE_1_SHORT_PER_WORD;
		}
	}
	else if (targetType == K_DATA_TYPE_2_SHORTS_PER_WORD)
	{
		if ((SortedDataBuffer.Type == K_DATA_TYPE_1_SHORT_PER_WORD) || (SortedDataBuffer.Type == K_DATA_TYPE_1_SHORT_PER_WORD_MAPPED_TO_POS))
		{
			SortedDataBuffer.SizeInWords = (SortedDataBuffer.SizeInWords + 1) >> 1;
			Put2ShortsIn1Word(SortedDataBuffer.Data, SortedDataBuffer.Data, SortedDataBuffer.SizeInWords);
			SortedDataBuffer.Type = K_DATA_TYPE_2_SHORTS_PER_WORD;
		}
	}
	
}


void LosslessCompressionForPhoto(COMPRESSION_MODE mode) 
{ 
    u4 NewSize, *swap; 
	u4 Size;
	
	CompressedDataBuffer.Data = ProcBuff + 1;
	swap = DramBuffer;
	
	/* Zero Repetition Suppression */
	NewSize= ZeroRepetitionSuppression(SortedDataBuffer.Data, SortedDataBuffer.SizeInWords, swap,0);
	MemCopyDmToDm(swap, SortedDataBuffer.Data, NewSize);
	SortedDataBuffer.SizeInWords = NewSize;
	
	//all compression algo needs the same format here
	ConvertSortedDataBuffer(K_DATA_TYPE_2_SHORTS_PER_WORD);
	
	if (mode.BolcDataType == 1) /* Test Pattern */
	{
		NewSize = (RZip32(SortedDataBuffer.Data, NewSize*4, CompressedDataBuffer.Data + 1, swap, 7)+3)>>2;
		CompressedDataBuffer.Data[1] |= 0x70000000;
		CompressedDataBuffer.Data[0] = 0x00000000;
		Size=1;
	}
	else
	{
		switch(DetectorConstantsPhoto.LosslessCompressionAlgorithm)
		{
		case 0:  /* Pacs_codec + Arithmetic Coder*/
			{
				if(NewSize>0x7FFF)
				{
					/* Arithmetic Coding */
					NewSize = (arit_cod(SortedDataBuffer.Data, NewSize*4, CompressedDataBuffer.Data + 1, swap)+3)>>2;
					Size=1;
					CompressedDataBuffer.Data[0] = 0xAA000000;
				}
				else
				{
					Size=NewSize>>1;
					
					/* Pacs Codec */
					NewSize=new_srt(SortedDataBuffer.Data+Size, NewSize-Size, CompressedDataBuffer.Data + 1, DetectorConstantsPhoto.PacsCodecOrder)+2;
					
					/* Arithmetic Coding */
					Size = (arit_cod(SortedDataBuffer.Data, Size*4, CompressedDataBuffer.Data + 1 + NewSize, swap)+7)>>2;
					
					CompressedDataBuffer.Data[0] = (0xCCAA0000|NewSize);
				}
			};break;
		case 1: /* Pacs_codec + Rzip 5*/
			{
				if(NewSize>0x7FFF)
				{
					/* Arithmetic Coding */
					NewSize = (arit_cod(SortedDataBuffer.Data, NewSize*4, CompressedDataBuffer.Data + 1, swap)+3)>>2;
					Size=1;
					CompressedDataBuffer.Data[0] = 0xAA000000;
				}
				else
				{
					Size=NewSize>>1;
					
					/* Pacs Codec */
					NewSize=new_srt(SortedDataBuffer.Data+Size, NewSize-Size, CompressedDataBuffer.Data + 1, DetectorConstantsPhoto.PacsCodecOrder)+2;
					
					/* RZip 5 */
					Size = (RZip32(SortedDataBuffer.Data, Size*4, CompressedDataBuffer.Data + 1 + NewSize, swap, 5)+7)>>2;
					CompressedDataBuffer.Data[NewSize+1] |= 0x50000000;
					
					CompressedDataBuffer.Data[0] = (0xCC000000|NewSize);
				}
			};break;
		case 2: /* Pacs_codec + Rzip 6*/
			{
				if(NewSize>0x7FFF)
				{
					/* Arithmetic Coding */
					NewSize = (arit_cod(SortedDataBuffer.Data, NewSize*4, CompressedDataBuffer.Data + 1, swap)+3)>>2;
					Size=1;
					CompressedDataBuffer.Data[0] = 0xAA000000;
				}
				else
				{
					Size=NewSize>>1;
					
					/* Pacs Codec */
					NewSize=new_srt(SortedDataBuffer.Data+Size, NewSize-Size, CompressedDataBuffer.Data + 1, DetectorConstantsPhoto.PacsCodecOrder)+2;
					
					/* RZip 5 */
					Size = (RZip32(SortedDataBuffer.Data, Size*4, CompressedDataBuffer.Data + 1 + NewSize, swap, 6)+7)>>2;
					CompressedDataBuffer.Data[NewSize+1] |= 0x60000000;
					
					CompressedDataBuffer.Data[0]= (0xCC000000|NewSize);
				}
			};break;
		case 3: /* Rzip */
			{
				/* RZip 6 */
				NewSize = (RZip32(SortedDataBuffer.Data, NewSize*4, CompressedDataBuffer.Data + 1, swap,6)+3)>>2;
				CompressedDataBuffer.Data[1] |= 0x60000000;
				Size=1;
				CompressedDataBuffer.Data[0] = 0x00000000;
			};break;
		case 4:	/* Arithmetic Coding */
			{
				/* Arithmetic Coding */
				NewSize = (arit_cod(SortedDataBuffer.Data, NewSize*4, CompressedDataBuffer.Data + 1, swap)+3)>>2;
				Size=1;
				CompressedDataBuffer.Data[0]= 0xAA000000;
			};break;
		default: /* Rzip 6 */
			{
				Result.LlcError |= (DetectorConstantsPhoto.LosslessCompressionAlgorithm | 0x80);
				NewSize = (RZip32(SortedDataBuffer.Data, NewSize*4, CompressedDataBuffer.Data +1, swap, 6)+3)>>2;
				CompressedDataBuffer.Data[1] |= 0x60000000;
				Size=1;
				CompressedDataBuffer.Data[0]= 0x00000000;
			};break;
		}
	}
	
	CompressedDataBuffer.SizeInWords = NewSize+Size;
} 

void LosslessCompressionForSpectro(COMPRESSION_MODE mode) 
{ 
    u4 NewSize1,*swap, key1, key2, dist1, dist2; 
	
	CompressedDataBuffer.Data = ProcBuff + 1;
	swap = DramBuffer;
	
	/* Adapt the distance to the RRR*/
	if (mode.ReadoutsPerRamps<=64)
	{
		dist1=6;
		dist2=4;
		key1= 0x60000000;
		key2= 0x64000000;
	}
	else
	{
		dist1=4;
		dist2=7;
		key1= 0x40000000;
		key2= 0x47000000;
	}
	
	switch(DetectorConstantsSpectro.LosslessCompressionAlgorithm)
	{
	case 0:
		{
			ConvertSortedDataBuffer(K_DATA_TYPE_2_SHORTS_PER_WORD);
			
			if(SortedDataBuffer.SizeInWords > 0x3FFF)
			{
				NewSize1 = (RZip32(SortedDataBuffer.Data, SortedDataBuffer.SizeInWords*4, CompressedDataBuffer.Data, swap, 6)+3)>>2;
				CompressedDataBuffer.Data[0] |= 0x60000000;
			}
			else
			{
				NewSize1=new_srt(SortedDataBuffer.Data, SortedDataBuffer.SizeInWords, CompressedDataBuffer.Data, DetectorConstantsSpectro.PacsCodecOrder)+2;
			}
		};break;
	case 1:
		{
			ConvertSortedDataBuffer(K_DATA_TYPE_2_SHORTS_PER_WORD);
			
			//first, we compress the data to ProcBuff
			NewSize1 = (RZip32(SortedDataBuffer.Data, SortedDataBuffer.SizeInWords*4, ProcBuff, swap, dist1)+3)>>2;
			ProcBuff[0] |= key1;
			
			//then, we compress the data to the nominal buffer
			if(SortedDataBuffer.SizeInWords > 0x3FFF)
			{
				NewSize1 = (RZip32(ProcBuff, NewSize1*4, CompressedDataBuffer.Data, swap, dist1)+3)>>2;
				CompressedDataBuffer.Data[0] |= key1;
			}
			else
			{
				NewSize1=new_srt(ProcBuff, NewSize1, CompressedDataBuffer.Data, DetectorConstantsSpectro.PacsCodecOrder)+2;
			}
		};break;
	case 2:
		{
			ConvertSortedDataBuffer(K_DATA_TYPE_2_SHORTS_PER_WORD);
			
			NewSize1 = (RZip32(SortedDataBuffer.Data, SortedDataBuffer.SizeInWords*4, CompressedDataBuffer.Data, swap, dist1)+3)>>2;
			CompressedDataBuffer.Data[0] |= key1;
		};break;
	case 3:
		{
			NewSize1 = (arit_cod(SortedDataBuffer.Data, SortedDataBuffer.SizeInWords*4, CompressedDataBuffer.Data, swap)+3)>>2;
		};break;
	case 4:
		{
			ConvertSortedDataBuffer(K_DATA_TYPE_1_SHORT_PER_WORD);
			
			NewSize1 = fm_ariencode(SortedDataBuffer.Data, SortedDataBuffer.SizeInWords, CompressedDataBuffer.Data, swap, SPEC_MODEL);
		};break;
	default:
		{
			Result.LlcError |= (DetectorConstantsSpectro.LosslessCompressionAlgorithm | 0x80);
			if(SortedDataBuffer.SizeInWords > 0x3FFF)
			{
				NewSize1 = (RZip32(SortedDataBuffer.Data, SortedDataBuffer.SizeInWords*4, CompressedDataBuffer.Data, swap, 6)+3)>>2;
				CompressedDataBuffer.Data[0] |= 0x60000000;
			}
			else
			{
				NewSize1=new_srt(SortedDataBuffer.Data, SortedDataBuffer.SizeInWords, CompressedDataBuffer.Data, DetectorConstantsSpectro.PacsCodecOrder)+2;
			}
		};break;
	} 
	
	/* Write the New Size in Words after LLC */
	CompressedDataBuffer.SizeInWords = NewSize1;
} 

/*********************************************************************************************************************
NEW FM ARI
*******************************************************************************************************************/


#define FMARIROWS 256
#define FMARISPILL 257

#define FM_ARI_MAXFREQ 8192
//#define MAXFREQ 16380

#define SPILLCUT FMARIROWS

u4 fm_ari_low = 0;
u4 fm_ari_high = 0xffff;
u4 fm_ari_underflow = 0;
u4 fm_ari_wctr = 0; 
u4 fm_ari_rctr = 0;



void init_arimodel (struct arimodel *model, u4 *buffer, int symbols)
{   // note: symbols is counted here without the spill probability
    model->freqtable = buffer; 
    model->cptable = model->freqtable + symbols + 1; // cumulative probability table 
    model->ncptable = model->cptable + 1; // next cumulative probability table
    model->spillover = model->ncptable + symbols + 1; // swap buffer for spillover, i.e. words > 8 Bit   
    model->spillctr = 0;
    model->probability = 0;
	
    return;
}

void encodeSym8k (BIT_STREAM* bitStream, u4 cp, u4 ncp, int probability)
{
	u4 width;
	u4 a;
	
	// calculate the new interval
	width = (fm_ari_high - fm_ari_low) + 1;
	// 
	fm_ari_high = fm_ari_low + ((ncp * width) >> 13) - 1; // L + Pni * (H - L)
	//
	fm_ari_low = fm_ari_low + ((cp * width) >> 13); // L + Pci * (H - L)
	
	for ( ; ; )
	{
		a = fm_ari_high & 0x8000;
		
		// write out equal bits
		if (a == (fm_ari_low & 0x8000))
		{
			PushBits(*bitStream, a >> 15, 1);
			fm_ari_wctr++;
			//putbit32 (dest, fm_ari_wctr++, a >> 15);
			
			while (fm_ari_underflow > 0)
			{
				PushBits(*bitStream, (~fm_ari_high & 0x8000)>>15, 1);
				fm_ari_wctr++;
				//putbit32 (dest, fm_ari_wctr++, (~fm_ari_high & 0x8000)>>15);
				fm_ari_underflow--;
			}
		}
		
		// underflow coming up, because <> and the 2nd bits are just one apart      
		else if ((fm_ari_low & 0x4000) && !(fm_ari_high & 0x4000))
		{
			fm_ari_underflow++;
			fm_ari_low &= 0x3fff;
			fm_ari_high |= 0x4000;
		}
		else
		{
			return ;
		}
		
		fm_ari_low <<= 1;
		fm_ari_low &= 0xffff;
		fm_ari_high <<= 1;
		fm_ari_high |= 1;
		fm_ari_high &= 0xffff;
	}
	return;
}


void fm_flushEncoder (BIT_STREAM* bitStream)
{
	PushBits(*bitStream, (fm_ari_low & 0x4000)>>14, 1);
	//putbit32 (dest, fm_ari_wctr++, (fm_ari_low & 0x4000)>>14);
	fm_ari_wctr++;
	
	fm_ari_underflow++;
	
	while (fm_ari_underflow-- > 0)
	{
		PushBits(*bitStream, (~fm_ari_low & 0x4000)>>14, 1);
		fm_ari_wctr++;
		//putbit32 (dest, fm_ari_wctr++, (~fm_ari_low & 0x4000)>>14);
	}
	
	return;
}


void InitModelTable (u4 *destmodel, int tableid)
{
    if (tableid == PHOT_MODEL)
    {
		// startmodel for default photometry
		destmodel[0] = 440;	destmodel[1] = 440;	destmodel[2]  = 425;	destmodel[3]  = 415;
		destmodel[4]  = 410; 	destmodel[5]  = 390;	destmodel[6]  = 390;	destmodel[7]  = 350;
		destmodel[8]  = 330;	destmodel[9]  = 300;	destmodel[10] = 295;	destmodel[11] = 255;
		destmodel[12] = 250;	destmodel[13] = 210;	destmodel[14] = 205;	destmodel[15] = 175;
		destmodel[16] = 175;	destmodel[17] = 140;	destmodel[18] = 140;	destmodel[19] = 110;
		destmodel[20] = 110;	destmodel[21] = 85;	destmodel[22] = 90;	destmodel[23] = 75;
		destmodel[24] = 70;	destmodel[25] = 55;	destmodel[26] = 50;	destmodel[27] = 45;
		destmodel[28] = 40;	destmodel[29] = 40;	destmodel[30] = 35;	destmodel[31] = 35;
		destmodel[32] = 30;	destmodel[33] = 25;	destmodel[34] = 25;	destmodel[35] = 20;
		destmodel[36] = 20;	destmodel[37] = 20;	destmodel[38] = 17;	destmodel[39] = 17;
		destmodel[40] = 17;	destmodel[41] = 15;	destmodel[42] = 15;	destmodel[43] = 15;
		destmodel[44] = 13;	destmodel[45] = 13;	destmodel[46] = 13;	destmodel[47] = 11;
		destmodel[48] = 11;	destmodel[49] = 11;	destmodel[50] = 10;	destmodel[51] = 10;
		destmodel[52] = 10;	destmodel[53] = 9;	destmodel[54] = 9;	destmodel[55] = 9;
		destmodel[56] = 8;	destmodel[57] = 8;	destmodel[58] = 8;	destmodel[59] = 8;
		destmodel[60] = 7;	destmodel[61] = 7;	destmodel[62] = 7;	destmodel[63] = 7;
		destmodel[64] = 6;	destmodel[65] = 6;	destmodel[66] = 6;	destmodel[67] = 6;
		destmodel[68] = 5;	destmodel[69] = 5;	destmodel[70] = 5;	destmodel[71] = 5;
		destmodel[72] = 4;	destmodel[73] = 4;	destmodel[74] = 4;	destmodel[75] = 4;
		destmodel[76] = 3;	destmodel[77] = 3;	destmodel[78] = 3;	destmodel[79] = 3;
		destmodel[80] = 3;	destmodel[81] = 3;	destmodel[82] = 2;	destmodel[83] = 2;
		destmodel[84] = 2;	destmodel[85] = 2;	destmodel[86] = 2;	destmodel[87] = 2;
		destmodel[88] = 1;	destmodel[89] = 1;	destmodel[90] = 1;	destmodel[91] = 1;
		destmodel[92] = 1;	destmodel[93] = 1;	destmodel[94] = 1;	destmodel[95] = 1;
		destmodel[96] = 1;	destmodel[97] = 1;	destmodel[98] = 1;	destmodel[99] = 1;
		destmodel[100] = 1;	destmodel[101] = 1;	destmodel[102] = 1;	destmodel[103] = 1;
		destmodel[104] = 1;	destmodel[105] = 1;	destmodel[106] = 1;	destmodel[107] = 1;
		destmodel[108] = 1;	destmodel[109] = 1;	destmodel[110] = 1;	destmodel[111] = 1;
		destmodel[112] = 1;	destmodel[113] = 1;	destmodel[114] = 1;	destmodel[115] = 1;
		destmodel[116] = 1;	destmodel[117] = 1;	destmodel[118] = 1;	destmodel[119] = 1;
		destmodel[120] = 1;	destmodel[121] = 1;	destmodel[122] = 1;	destmodel[123] = 1;
		destmodel[124] = 1;	destmodel[125] = 1;	destmodel[126] = 1;	destmodel[127] = 1;
		destmodel[128] = 1;	destmodel[129] = 1;	destmodel[130] = 1;	destmodel[131] = 1;
		destmodel[132] = 1;	destmodel[133] = 1;	destmodel[134] = 1;	destmodel[135] = 1;
		destmodel[136] = 1;	destmodel[137] = 1;	destmodel[138] = 1;	destmodel[139] = 1;
		destmodel[140] = 1;	destmodel[141] = 1;	destmodel[142] = 1;	destmodel[143] = 1;
		destmodel[144] = 1;	destmodel[145] = 1;	destmodel[146] = 1;	destmodel[147] = 1;
		destmodel[148] = 1;	destmodel[149] = 1;	destmodel[150] = 1;	destmodel[151] = 1;
		destmodel[152] = 1;	destmodel[153] = 1;	destmodel[154] = 1;	destmodel[155] = 1;
		destmodel[156] = 1;	destmodel[157] = 1;	destmodel[158] = 1;	destmodel[159] = 1;
		destmodel[160] = 1;	destmodel[161] = 1;	destmodel[162] = 1;	destmodel[163] = 1;
		destmodel[164] = 1;	destmodel[165] = 1;	destmodel[166] = 1;	destmodel[167] = 1;
		destmodel[168] = 1;	destmodel[169] = 1;	destmodel[170] = 1;	destmodel[171] = 1;
		destmodel[172] = 1;	destmodel[173] = 1;	destmodel[174] = 1;	destmodel[175] = 1;
		destmodel[176] = 1;	destmodel[177] = 1;	destmodel[178] = 1;	destmodel[179] = 1;
		destmodel[180] = 1;	destmodel[181] = 1;	destmodel[182] = 1;	destmodel[183] = 1;
		destmodel[184] = 1;	destmodel[185] = 1;	destmodel[186] = 1;	destmodel[187] = 1;
		destmodel[188] = 1;	destmodel[189] = 1;	destmodel[190] = 1;	destmodel[191] = 1;
		destmodel[192] = 1;	destmodel[193] = 1;	destmodel[194] = 1;	destmodel[195] = 1;
		destmodel[196] = 1;	destmodel[197] = 1;	destmodel[198] = 1;	destmodel[199] = 1;
		destmodel[200] = 1;	destmodel[201] = 1;	destmodel[202] = 1;	destmodel[203] = 1;
		destmodel[204] = 1;	destmodel[205] = 1;	destmodel[206] = 1;	destmodel[207] = 1;
		destmodel[208] = 1;	destmodel[209] = 1;	destmodel[210] = 1;	destmodel[211] = 1;
		destmodel[212] = 1;	destmodel[213] = 1;	destmodel[214] = 1;	destmodel[215] = 1;
		destmodel[216] = 1;	destmodel[217] = 1;	destmodel[218] = 1;	destmodel[219] = 1;
		destmodel[220] = 1;	destmodel[221] = 1;	destmodel[222] = 1;	destmodel[223] = 1;
		destmodel[224] = 1;	destmodel[225] = 1;	destmodel[226] = 1;	destmodel[227] = 1;
		destmodel[228] = 1;	destmodel[229] = 1;	destmodel[230] = 1;	destmodel[231] = 1;
		destmodel[232] = 1;	destmodel[233] = 1;	destmodel[234] = 1;	destmodel[235] = 1;
		destmodel[236] = 1;	destmodel[237] = 1;	destmodel[238] = 1;	destmodel[239] = 1;
		destmodel[240] = 1;	destmodel[241] = 1;	destmodel[242] = 1;	destmodel[243] = 1;
		destmodel[244] = 1;	destmodel[245] = 1;	destmodel[246] = 1;	destmodel[247] = 1;
		destmodel[248] = 1;	destmodel[249] = 1;	destmodel[250] = 1;	destmodel[251] = 1;
		destmodel[252] = 1;	destmodel[253] = 1;	destmodel[254] = 1;	destmodel[255] = 1;
		destmodel[256] = 934; // spillover
    }
    else if (tableid == SPEC_MODEL)
    {
		// startmodel for default spectroscopy
		destmodel[0] = 1400;	destmodel[1] = 1300;	destmodel[2]  = 1300;	destmodel[3]  = 900;
		destmodel[4]  = 900; 	destmodel[5]  = 500;	destmodel[6]  = 500;	destmodel[7]  = 200;
		destmodel[8]  = 180;	destmodel[9]  = 70;	destmodel[10] = 60;	destmodel[11] = 30;
		destmodel[12] = 20;	destmodel[13] = 10;	destmodel[14] = 10;	destmodel[15] = 10;
		destmodel[16] = 5;	destmodel[17] = 5;	destmodel[18] = 5;	destmodel[19] = 5;
		destmodel[20] = 5;	destmodel[21] = 5;	destmodel[22] = 5;	destmodel[23] = 5;
		destmodel[24] = 5;	destmodel[25] = 5;	destmodel[26] = 5;	destmodel[27] = 5;
		destmodel[28] = 5;	destmodel[29] = 5;	destmodel[30] = 5;	destmodel[31] = 5;
		destmodel[32] = 2;	destmodel[33] = 2;	destmodel[34] = 2;	destmodel[35] = 2;
		destmodel[36] = 2;	destmodel[37] = 2;	destmodel[38] = 2;	destmodel[39] = 2;
		destmodel[40] = 2;	destmodel[41] = 2;	destmodel[42] = 2;	destmodel[43] = 2;
		destmodel[44] = 2;	destmodel[45] = 2;	destmodel[46] = 2;	destmodel[47] = 2;
		destmodel[48] = 2;	destmodel[49] = 2;	destmodel[50] = 2;	destmodel[51] = 2;
		destmodel[52] = 2;	destmodel[53] = 2;	destmodel[54] = 2;	destmodel[55] = 2;
		destmodel[56] = 2;	destmodel[57] = 2;	destmodel[58] = 2;	destmodel[59] = 2;
		destmodel[60] = 2;	destmodel[61] = 2;	destmodel[62] = 2;	destmodel[63] = 2;
		destmodel[64] = 1;	destmodel[65] = 1;	destmodel[66] = 1;	destmodel[67] = 1;
		destmodel[68] = 1;	destmodel[69] = 1;	destmodel[70] = 1;	destmodel[71] = 1;
		destmodel[72] = 1;	destmodel[73] = 1;	destmodel[74] = 1;	destmodel[75] = 1;
		destmodel[76] = 1;	destmodel[77] = 1;	destmodel[78] = 1;	destmodel[79] = 1;
		destmodel[80] = 1;	destmodel[81] = 1;	destmodel[82] = 1;	destmodel[83] = 1;
		destmodel[84] = 1;	destmodel[85] = 1;	destmodel[86] = 1;	destmodel[87] = 1;
		destmodel[88] = 1;	destmodel[89] = 1;	destmodel[90] = 1;	destmodel[91] = 1;
		destmodel[92] = 1;	destmodel[93] = 1;	destmodel[94] = 1;	destmodel[95] = 1;
		destmodel[96] = 1;	destmodel[97] = 1;	destmodel[98] = 1;	destmodel[99] = 1;
		destmodel[100] = 1;	destmodel[101] = 1;	destmodel[102] = 1;	destmodel[103] = 1;
		destmodel[104] = 1;	destmodel[105] = 1;	destmodel[106] = 1;	destmodel[107] = 1;
		destmodel[108] = 1;	destmodel[109] = 1;	destmodel[110] = 1;	destmodel[111] = 1;
		destmodel[112] = 1;	destmodel[113] = 1;	destmodel[114] = 1;	destmodel[115] = 1;
		destmodel[116] = 1;	destmodel[117] = 1;	destmodel[118] = 1;	destmodel[119] = 1;
		destmodel[120] = 1;	destmodel[121] = 1;	destmodel[122] = 1;	destmodel[123] = 1;
		destmodel[124] = 1;	destmodel[125] = 1;	destmodel[126] = 1;	destmodel[127] = 1;
		destmodel[128] = 1;	destmodel[129] = 1;	destmodel[130] = 1;	destmodel[131] = 1;
		destmodel[132] = 1;	destmodel[133] = 1;	destmodel[134] = 1;	destmodel[135] = 1;
		destmodel[136] = 1;	destmodel[137] = 1;	destmodel[138] = 1;	destmodel[139] = 1;
		destmodel[140] = 1;	destmodel[141] = 1;	destmodel[142] = 1;	destmodel[143] = 1;
		destmodel[144] = 1;	destmodel[145] = 1;	destmodel[146] = 1;	destmodel[147] = 1;
		destmodel[148] = 1;	destmodel[149] = 1;	destmodel[150] = 1;	destmodel[151] = 1;
		destmodel[152] = 1;	destmodel[153] = 1;	destmodel[154] = 1;	destmodel[155] = 1;
		destmodel[156] = 1;	destmodel[157] = 1;	destmodel[158] = 1;	destmodel[159] = 1;
		destmodel[160] = 1;	destmodel[161] = 1;	destmodel[162] = 1;	destmodel[163] = 1;
		destmodel[164] = 1;	destmodel[165] = 1;	destmodel[166] = 1;	destmodel[167] = 1;
		destmodel[168] = 1;	destmodel[169] = 1;	destmodel[170] = 1;	destmodel[171] = 1;
		destmodel[172] = 1;	destmodel[173] = 1;	destmodel[174] = 1;	destmodel[175] = 1;
		destmodel[176] = 1;	destmodel[177] = 1;	destmodel[178] = 1;	destmodel[179] = 1;
		destmodel[180] = 1;	destmodel[181] = 1;	destmodel[182] = 1;	destmodel[183] = 1;
		destmodel[184] = 1;	destmodel[185] = 1;	destmodel[186] = 1;	destmodel[187] = 1;
		destmodel[188] = 1;	destmodel[189] = 1;	destmodel[190] = 1;	destmodel[191] = 1;
		destmodel[192] = 1;	destmodel[193] = 1;	destmodel[194] = 1;	destmodel[195] = 1;
		destmodel[196] = 1;	destmodel[197] = 1;	destmodel[198] = 1;	destmodel[199] = 1;
		destmodel[200] = 1;	destmodel[201] = 1;	destmodel[202] = 1;	destmodel[203] = 1;
		destmodel[204] = 1;	destmodel[205] = 1;	destmodel[206] = 1;	destmodel[207] = 1;
		destmodel[208] = 1;	destmodel[209] = 1;	destmodel[210] = 1;	destmodel[211] = 1;
		destmodel[212] = 1;	destmodel[213] = 1;	destmodel[214] = 1;	destmodel[215] = 1;
		destmodel[216] = 1;	destmodel[217] = 1;	destmodel[218] = 1;	destmodel[219] = 1;
		destmodel[220] = 1;	destmodel[221] = 1;	destmodel[222] = 1;	destmodel[223] = 1;
		destmodel[224] = 1;	destmodel[225] = 1;	destmodel[226] = 1;	destmodel[227] = 1;
		destmodel[228] = 1;	destmodel[229] = 1;	destmodel[230] = 1;	destmodel[231] = 1;
		destmodel[232] = 1;	destmodel[233] = 1;	destmodel[234] = 1;	destmodel[235] = 1;
		destmodel[236] = 1;	destmodel[237] = 1;	destmodel[238] = 1;	destmodel[239] = 1;
		destmodel[240] = 1;	destmodel[241] = 1;	destmodel[242] = 1;	destmodel[243] = 1;
		destmodel[244] = 1;	destmodel[245] = 1;	destmodel[246] = 1;	destmodel[247] = 1;
		destmodel[248] = 1;	destmodel[249] = 1;	destmodel[250] = 1;	destmodel[251] = 1;
		destmodel[252] = 1;	destmodel[253] = 1;	destmodel[254] = 1;	destmodel[255] = 1;
		destmodel[256] = 466; // spillover
    }
	
    return;
}


#define DUMPMODEL(m) { int c; for (c=0; c < FMARISPILL; c++) printf ("i: %d p: %d cp: %d ncp: %d\n", c, m[c], m[c+FMARISPILL], m[c+FMARISPILL+1]); }
#define CHECKMODEL(m) { int v, c; for (v=0, c=0; c < FMARISPILL; c++) v += m[c]; /*printf ("Model sum is: %d\n", v);*/ }

void update_fm_ari_model (u4 *buffer, u4 nwords, int symbols, int initval, struct arimodel *model)
{
    unsigned int ctr;
    int value;
	
    // start model with 0 or 1 in every entry -> smooth
    for (ctr=0; ctr < FMARISPILL; ctr++) 
	{
		model->freqtable[ctr] = initval;
	}
    // count freqs over the buffer, but leave out the first FMACROWS words for smoothing  
    for (ctr=symbols; ctr < nwords; ctr++)
    {
		value = buffer[ctr]; 
		//printf ("updatemodel [%d] = %d\n", ctr, buffer[ctr]);
		
		if (value < FMARIROWS)
		{
			model->freqtable[value]++;
		}
		else
		{
			model->freqtable[FMARIROWS]++; // spillover
		}
    }	
	
    // just checking... 
    CHECKMODEL(model->freqtable);
	
    // make new (n)cp array
    model->probability = makeCumulative (model->freqtable, FMARISPILL, model->cptable); 
    
    return;
}


int fm_encode_chunk (u4 *chunk, int chunksize, unsigned int *dest, struct arimodel *model)
{
    int ctr, tail;
    u4 symbol, cp, ncp;
	
	BIT_STREAM bitStream;
	
	DeclareBitStream(bitStream, dest);
	PushBits(bitStream, 0, 32);	//first cell will be used later to store the size
	
    // now init ari
    fm_ari_low = 0;
    fm_ari_high = 0xffff;
    fm_ari_underflow = 0;
    fm_ari_wctr = 32; // offset for chunksize_w
	
	//   DUMPMODEL(model->freqtable);
    CHECKMODEL(model->freqtable);
	
    for (ctr=0; ctr < chunksize; ctr++)
    {
		// get next symbol
		symbol = chunk[ctr];
		
		//printf("the symbol you check out is: chunk[%d]=%d\n", ctr, symbol);
		// look it up in the tables
		
		// first we check for spillover
		if (symbol >= SPILLCUT) 
		{
			// encode spillover signal in ari stream
			cp = model->cptable[FMARIROWS];
			ncp = model->ncptable[FMARIROWS];
			//
			// printf ("we will put symbol %d found at %d aside (%d)!\n", symbol, ctr, model[FMARIROWS]);
			//
			encodeSym8k (&bitStream, cp, ncp, model->probability);
			
			// put the symbol into the spillover buffer and increase counter	    
			// printf ("hauts di hier auf? spillctr=%d\n", *spillctr);
			model->spillover[(model->spillctr)++] = symbol;
		}
		else // valid symbol
		{	    
			cp = model->cptable[symbol];
			ncp = model->ncptable[symbol];
			
			encodeSym8k (&bitStream, cp, ncp, model->probability);
		}
		
		//printf("just encoded %lx, wctr is %ld\n", symbol, ari_wctr);
    }
    // encode the rest
	
    fm_flushEncoder (&bitStream);
	
    // calc fillup and fill up with 0s
    tail = (32 - bitStream.CurrentBitPos % 32) % 32; 
    fm_ari_wctr += tail; 
    dest[0] = (bitStream.CurrentBitPos + tail) >> 5;
	PadBitStreamWithZeroes(bitStream);
	dest[0] = bitStream.CurrentBitPos >> 5;
	
    return dest[0]; // now in words 
}

int vbwl_midsize (u4 *source, int words, unsigned int *dest, int BS)
{
    int ctr, bctr;
    int bits, width;
    int outbits = 32; // keep track of the output bits; we start at dest[1]
	
	BIT_STREAM bitStream;
	
	DeclareBitStream(bitStream, dest);
	PushBits(bitStream, 0, 32);	//keep first cell for later use
	
#define VBWLCODE 3
#define VBWLMINW 9
	
    // main loop counts through the source words
    for (ctr=0; ctr < words; ctr++)
    {
        //printf ("hi, ctr is %d\n", ctr);
        // block-loop, we count through the words of a block
        for (width=0, bctr=ctr; (bctr < ctr+BS) & (bctr < words); bctr++)
        {
            // determine used bits of current word
            // bits = 32-lzeros
            for (bits=1; (source[bctr] >> bits) > 0; bits++); // start with 1, otherwise the word 0 would use 0 bit
			{
				//	    printf ("symbol: %d bits=%d\n", source[bctr], bits);
				width = bits > width ? bits : width;
				//          printf ("width=%d\n",width);
			}
        }
        // now we know width = the number of bits to encode the block
		
        // first code the width
		if (width < 9) // ... due to the optional -FMARIROWS
		{
			width = 9;
		}
		else 
		{
			if (width > 16)
			{
				//	printf ("ERROR at ctr=%d, word is %x and width is %d!\n",ctr,source[ctr], width);
			}
		}
        //printf ("code width %d(%d) to %d\n", width-VBWLMINW, width, outbits);
        // decide whether to use 3 or (3+)4 bits for encoding the width
		
		PushBits(bitStream, width - VBWLMINW, VBWLCODE); // -9 because: 9...16 -> 0...7
		outbits += VBWLCODE;
		
        // now code the words of the block
        for (bctr=ctr; (ctr < bctr+BS) & (ctr < words); ctr++)
        {
            //printf ("code %d now to %d\n", source[ctr], outbits);
			PushBits(bitStream, source[ctr], width);
            outbits += width;
        }
        ctr--;
    }
	
    // store the original size
    dest[0] = words;
	
    // return the size in words, rounding up
    return (outbits+31) >> 5;
}



// input data should be 16 bit values stored in 32 bit words 
// swap should allow up to nwords+257+258 words, but 1/10 should be sufficient
int fm_ariencode (u4 *source, u4 nwords, u4 *dest, u4 *swap, int modelId)
{
    int ctr, src_ctr, remaining; 
	
    u4 *streamOfChunks;
    unsigned int *streamctr_w;
	//   int chunksize;
	
    struct arimodel model;
	
	//    int symbol, cp, ncp; // cp = cumulative probability, ncp = next cumulative probability
	
#define ARIHDR 2
	
    init_arimodel (&model, swap, FMARIROWS);
    
    // store original size in words
    dest[0] = nwords; // original size in words
    remaining = nwords;
    src_ctr = 0;
    streamctr_w = dest + 1; // here the size of the ari stream in words will grow
    *streamctr_w = 0; // we start with 2, because we have the osize and the streamctr
	
    // set dest stream and counter
    streamOfChunks = dest + ARIHDR;
	
    // initialize startmodel
    InitModelTable(model.freqtable, modelId);
    
    // make probability model
    model.probability = makeCumulative (model.freqtable, FMARISPILL, model.cptable);    
	
    // start compressing chunks with initial model    
    while (remaining > FM_ARI_MAXFREQ)
    {
		CHECKMODEL(model.freqtable);
		*streamctr_w += fm_encode_chunk (source + src_ctr, FM_ARI_MAXFREQ, streamOfChunks + *streamctr_w, &model);
		
		// derive new model from current data
		update_fm_ari_model (source + src_ctr, FM_ARI_MAXFREQ, FMARISPILL, 1, &model);	
		
		src_ctr += FM_ARI_MAXFREQ;
		remaining -= FM_ARI_MAXFREQ;
    }
	
    // encode the last chunk
    if (remaining > 0)
	{
		*streamctr_w += fm_encode_chunk (source + src_ctr, remaining, streamOfChunks + *streamctr_w, &model);
	}  
    //printf ("spillctr is in the end: %d\n", model.spillctr);
    // .. treat the spillover here
	
    // subtract FMARIROWS from the spill values
    for (ctr=0; ctr < model.spillctr; ctr++)
	{
		model.spillover[ctr] -= FMARIROWS;
	}
	
    // printout
	//    for (ctr=0; ctr < model.spillctr; ctr++) printf ("spill[%d]: %d\n", ctr, model.spillover[ctr]);
	
    model.spillctr = vbwl_midsize (model.spillover, model.spillctr, dest + ARIHDR + *streamctr_w, 4);
	
	//    printf ("ari stream: %d, spill: %d\n", *streamctr_w, model.spillctr);
	
	//    printf ("1st word of spill is osize: %d\n", *(dest + ARIHDR + *streamctr_w));
	
    return (int)(ARIHDR + *streamctr_w + model.spillctr); 
}


#ifdef SIMULATOR
void putnbits32(u4 value, u4 bitOffset, u4 nbits, u4 *dest)
{ 
	/* ok-but-slow-version*/
	u4 A, A1, A2,A3,offs=bitOffset>>5, *dest1;
	dest1 =dest+offs;
	
	A1 = (bitOffset&0x1f); /* end of the bitstring with local offset, overflowed bits: A-32*/
	A = A1+nbits; /* end of the bitstring with local offset, overflowed bits: A-32*/
	if (A <= 32)
    {
		*(dest1) &= WordMaskRight[A1];  /* clear destination */
		*(dest1) |= (value<<(32-A));  /* assign value */
    }
	else
    {
		A2 =32-A1;
		A3=A-32;
		putnbits32(value >> A3, bitOffset, A2, dest); /* process first part*/
		putnbits32(value, bitOffset+A2, A3, dest); /* process second part */
    }
}

void putnbits (u4 value, u4 offset, u4 nbits, u4 *dest)
{
	putnbits32(value, offset, nbits, dest);
}
#endif

