/***************************************************************************************
* File name: p_proc.c
* Purpose:  pre-Processes the Data.	 
*
* Description:	The Preprocessing Program is Performed to Reorder/Normalize the Data before
*				the Compression.
****************************************************************************************/

#ifdef SIMULATOR
#include "virtuosoSim.h"
#else // SIMULATOR
#include "node1.h"
#include "k_struct.h"
#endif
#include "spuAsw.h"
#include "spulib.h"


/*
Apply a pre-processing computation to the data before reduction and compression
*/
void PreProcessing(u4 detectorType)
{
	u4 operation = 0;
	u4 parameter = 0;
	u4 size = ProcBuff[0]; /* Read the Size of the Obtained Data Located in the Top Address */
	u4* buffer = ProcBuff+1;
	u4 i;
	
	/* Photometry or Spectroscopy Preprocessing */
	if(detectorType == K_SPECTRO) 
	{
		operation = DetectorConstantsSpectro.PreProcessingFlag;
		parameter = DetectorConstantsSpectro.PreProcessingParameter;
	}
	else if(detectorType == K_PHOTO) 
	{
		operation = DetectorConstantsPhoto.PreProcessingFlag;
		parameter = DetectorConstantsPhoto.PreProcessingParameter;
	}
	
	if (operation != 0)
	{
		switch (operation)
		{
			case 1:
			{
				/* Addition of the Offset to the Data */
				for(i=0;i<size;i++)
				{
					buffer[i] = buffer[i] + parameter;	
				}
			}break;
			case 2:
			{
				/* Substration of the Offset to the Data */
				for(i=0;i<size;i++)
				{
					buffer[i] = buffer[i] - parameter;	
				}
			}break;
			case 3:
			{
				/* Multiplication the Offset by Data Samples */
				for(i=0;i<size;i++)
				{
					buffer[i] = buffer[i] * parameter;	
				}
			}break;
			case 4:
			{
				/* Division the Data Sample by Offset */
				if (parameter != 0) 
				{
					for(i=0;i<size;i++)
					{
						buffer[i] = roldiv(buffer[i], parameter);	
					}
				}
			}break;
			default:
			return;
		}
	}
}

/* Configurable Frame Dropping as an optional pre-processing step */
/* included by RO in jan09 for Version 13.10 */
/* returns the number of values (u4 words) that are left after decimation */
/* works in place */

int DecimateFrameOrder (int *buffer, u4 framesize, u4 frames, u4 groupsize, u4 discard_pre, u4 discard_post)
{
    int read, write, gctr;
    int in = discard_pre - 1;
    int out = groupsize - discard_post;
	int z, roff, woff;
	int drop = discard_pre + discard_post;

    // check parameters, if they are not valid then do nothing
    if ((groupsize < drop) || (drop == 0))
    {
		write = frames;
    }
    else // decimate
    {
		read = 0; write = 0; gctr = 0;
		for ( ; read < frames; )
		{
	    	// prepare group counter
	    	if (gctr == groupsize) // if end of group
				gctr = 0; // reset group counter
	    
	    	if ((gctr > in) && (gctr < out)) // middle: copy and increase rw counters
	    	{
				// DO THE MEMCOPY memcopy from read*framesize to write*framesize					    
			    roff = read*framesize;
		    	woff = write*framesize;
		    	
		    	for (z=0; z<framesize; z++) 
					buffer[woff+z] = buffer[roff+z];
			
				read++;
				write++;
				gctr++;
	    	}		
	    	else // in the start or end part discard frame by increasing read counters  
	    	{
				read++;
				gctr++;
	    	}		
		}
    }
    
    return write;
}

// same for pixel order (the one actually used)
int DecimatePixelOrder (int *buffer, u4 framesize, u4 frames, u4 groupsize, u4 discard_pre, u4 discard_post)
{
    int gctr, write, framesleft;
    int in = discard_pre - 1;
    int out = groupsize - discard_post;
    int p, f; // pixel and frame counters
	int drop = discard_pre + discard_post;

    // check parameters, if they are not valid then do nothing
    if ((groupsize < drop) || (drop == 0))
    {
		write = frames*framesize;
    }
    else // decimate
    {
		write = 0; gctr = 0;
	
		for (p=0; p < framesize; p++) // count over all pixels
		{
	    	gctr=0;
	    	for (f=0; f < frames; f++) // count over all frames 
	    	{	    
				// prepare group counter
				if (gctr == groupsize) // if end of group
		    		gctr = 0; // reset group counter
	    
				if ((gctr > in) && (gctr < out)) // middle: copy and increase rw counters
				{
		    		buffer[write] = buffer[p*frames+f];
		
		    		write++;
		    		gctr++;
				}		
				else // in the start or end part discard frame by increasing read counters  
				{
		    		gctr++;
				}		
	    	}
		}
    }
	framesleft = roldiv(write,framesize);
    return framesleft;
}
